/******************************************************************************

  nsdperf - Test network performance, simulating GPFS NSD client / server

  See the README file for information about building and running
  this program.

  Changes in version 1.22:

    * Increase the maximum number of tester threads to 4096 and the
      maximum number of parallel connections to 8191.

    * Update help messages.

  Changes in version 1.20:

    * Command line option for specifying the number of receiver threads is
      now "-t" rather than "-r".

    * Add "-r" command line option for specifying the RDMA device and port.

    * On Linux systems, use epoll to wait for socket data.

  Changes in version 1.16:

    * Add a command to verify data message contents.

    * Add new RDMA options for sending control messages over the RDMA
      interface and sending data inline with control messages.

  Changes in version 1.14:

    * Add support for RDMA connections using verbs library.

    * Serialize writing of log message to stdout.

    * Include message rate in output statistics.

    * Require a space after "/" when used as a command delimiter.

  Changes in version 1.13:

    * Fixed ordering of in-use waiter queue that caused excessively
      long waits.

    * Use a separate pending reply table and message id for each
      connection to avoid contention for global locks.

    * Use nanosecond resolution for times.

******************************************************************************/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#define _LINUX_SOURCE_COMPAT

#include <algorithm>
#include <cstdio>
#include <cstdlib>
#include <deque>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <list>
#include <map>
#include <new>
#include <queue>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include <math.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/un.h>
#ifdef _AIX
#include <libperfstat.h>
#endif
#ifdef RDMA
#include <infiniband/verbs.h>
#endif

#ifdef AF_INET6
#define IPV6_SUPPORT
#endif

#ifdef __linux
#define USE_EPOLL
#endif

#ifdef USE_EPOLL
#include <sys/epoll.h>
#endif

#ifdef __sparc
#define strerror_r(_e,_b,_s) strerror(_e)
#define INADDR_NONE INADDR_BROADCAST
#include <sys/filio.h>
#endif

using namespace std;


// Sized types
typedef unsigned char          UChar;
typedef short                  Int16;
typedef int                    Int32;
typedef long long              Int64;
typedef unsigned short         UInt16;
typedef unsigned int           UInt32;
typedef unsigned long long     UInt64;

// High-resolution time (nanosecond units)
typedef long long HTime;

// These definitions allow compatibility with the Windows Socket Library.
// They work in conjunction with a GPFS interface library.
#ifndef USE_WINSOCK
  // Socket type
  typedef int Sock;

  // Socket event description for the poll() function
  typedef struct pollfd PollSock;

  // The customary invalid socket value.
  static const Sock INVALID_SOCK = -1;
#endif

// Message ID type
typedef UInt32 MsgId;

// Program version
static const string version = "1.22";

// Default port to use 
static const int NSDPERF_PORT = 6668;

// Default number of message worker threads
static const int MSG_WORKERS = 10;

// Default number of tester threads
static const unsigned int TESTER_THREADS = 4;

// Maximum number of tester threads
static const unsigned int MAX_TESTERS = 4096;

// Maximum size for data buffer
static const unsigned int MAX_BUFFSIZE = 8 * 1024 * 1024;

// Maximum size for data part of a control message
static const unsigned int MAX_RPCSIZE = 64;

// Maximum TCP send/receive buffer size
static const int MAX_SOCKSIZE = 100 * 1024 * 1024;

// Maximum number of FDs to pass to poll
static const int MAX_POLLFD_NUM = 8192;

// Maximum number of parallel connections
static const int MAX_PARALLEL = MAX_POLLFD_NUM - 1;

// Size of a message header
static const unsigned int MSG_HDRSIZE = 4 * sizeof(UInt32);

// Magic number for message headers
static const UInt32 MSG_MAGIC = 0x1F2E3D4CU;

// For scrambling data in test data buffers
static const UInt64 SCRAMBLE = 0x0305070b0d111317ULL;

// Maximum possible value for an HTime
static const long long MAX_HTIME = 0x7FFFFFFFFFFFFFFFLL;

// For ioctl arguments
static const int on = 1;

// Flag to use in sendmsg system call to prevent EPIPE errors.  Some
// systems don't have this.
#ifdef MSG_NOSIGNAL
static const int SENDMSG_FLAGS = MSG_NOSIGNAL;
#else
static const int SENDMSG_FLAGS = 0;
#endif


// Forward declarations
class MsgRecord;
class MsgWorker;
class PollWait;
struct RcvMsg;
class RdmaConn;
class Receiver;
struct Target;
struct TestReq;
class Tester;
class Thread;


// Error macros
#define Warn(msg) do \
  { cerr << progname << ": Warning: " << msg << endl; }\
  while (false)
#define Warnm(msg) do \
  { cerr << progname << ": Warning: " << msg << ": "; perror(""); }\
  while (false)
#define Error(msg) do \
  { cerr << progname << ": " << msg << endl; exit(EXIT_FAILURE); }\
  while (false)
#define Errorm(msg) do \
  { cerr << progname << ": " << msg << ": "; perror(""); exit(EXIT_FAILURE); }\
  while (false)

// Console log output macros
#define Log(msg) do \
  { thLock(&logMutex); cout << msg << endl; thUnlock(&logMutex); }\
  while (false)
#define Logt(msg) do \
  { thLock(&logMutex); \
    cout << getTime() << " " << msg << endl; \
    thUnlock(&logMutex); } while (false)
#define Logm(msg) do \
  { int e = errno; thLock(&logMutex); \
    cout << msg << ": " << geterr(e) << endl; \
    thUnlock(&logMutex); } while (false)


// Error codes
enum Errno
{
  E_OK,
  E_INVAL,
  E_NOENT,
  E_WOULDBLOCK,
  E_CONNRESET,
  E_BADMSG,
  E_BROKEN,
  E_SENDFAILED,
  E_REPLY,
  E_CONNFAILED
};


// Address types for IpAddr
enum Atype
{
  AT_IPV4,
  AT_IPV6
};

  
// Structure for holding an IPv4 or IPv6 address
struct IpAddr
{
  Atype fam;
  UChar a[16];
  IpAddr();
  bool operator<(const IpAddr &iaddr) const;
  bool operator!=(const IpAddr &iaddr) const;
  Errno parse(const string hostname);
  void loadSockaddr(const sockaddr *saddrP);
  string tostring() const;
#ifdef IPV6_SUPPORT
  int getFamily() const { return (fam == AT_IPV6) ? AF_INET6 : AF_INET; }
#else
  int getFamily() const { return AF_INET; }
#endif
};


// Structure for holding an RDMA memory address (i.e. a character pointer)
// for passing between nodes.
struct RdmaAddr
{
  UInt64 addr;
  RdmaAddr() { addr = 0; }
  RdmaAddr(UInt64 a) : addr(a) {}
  RdmaAddr(char *p) : addr(reinterpret_cast<UInt64>(p)) {}
  operator char *() { return reinterpret_cast<char *>(addr); }
  operator UInt64() { return addr; }
};


// Abstract base class for threads.  Derived class must supply the routine
// body.
class Thread
{
  pthread_t th;

protected:
  bool running;

public:
  virtual ~Thread() {}
  void init();
  void startup();
  void waitForDeath();
  pthread_t getThread() const { return th; }
  virtual int threadBody() = 0;
};


// Data buffer that allows putting and getting items for sending in messages
class DataBuff
{
  char *buffP;                  // The data buffer
  unsigned int alloc;           // Actual allocated length
  unsigned int bufflen;         // Length currently assigned to buffer
  unsigned int buffpos;         // Position for put/get
  char *auxBuffP;               // Extra buffer used for inline RDMA
  unsigned int auxlen;          // Length of extra buffer

public:
  DataBuff()
    { buffP = auxBuffP = NULL; alloc = bufflen = buffpos = auxlen = 0; }
  DataBuff(unsigned int len)           { alloc = 0; newBuff(len); }
  DataBuff(char *dP, unsigned int len) { alloc = 0; initBuff(dP, len); }
  ~DataBuff() { if (alloc > 0) delete [] buffP; }
  void newBuff(unsigned int len);
  void initBuff(char *dataP, unsigned int datalen);
  void fillBuff(UInt64 seed);
  bool verifyBuff(UInt64 seed);
  void resetBuff() { buffpos = 0; }
  char *getBuffP() { return buffP; }
  unsigned int getBufflen() const { return bufflen; }
  void setAux(char *dataP, unsigned int datalen)
    { auxBuffP = dataP; auxlen = datalen; }
  char *getAuxBuffP() { return auxBuffP; }
  unsigned int getAuxlen() { return auxlen; }
  void putUInt32(UInt32 i);
  void putUInt64(UInt64 i);
  void putRdmaAddr(RdmaAddr a);
  void putHTime(HTime t);
  void putIpAddr(IpAddr iaddr);
  void putString(string s);
  UInt32 getUInt32();
  UInt64 getUInt64();
  RdmaAddr getRdmaAddr();
  HTime getHTime();
  IpAddr getIpAddr();
  string getString();
};

  
// Gather histogram of response times
class Histogram
{
  map<HTime, UInt32> buckets;
  int nEvents;
  HTime totalTime;

public:
  Histogram() { nEvents = 0; totalTime = 0; }
  void addEntry(HTime t);
  void addHist(const Histogram *hP);
  void printHist(ostream &os) const;
  void putBuff(DataBuff *dbP) const;
  void getBuff(DataBuff *dbP);
  int getNevents() const { return nEvents; }
  unsigned int calcLen() const
    { return sizeof(UInt64) + 2 * sizeof(UInt32) +
        buckets.size() * (sizeof(UInt64) + sizeof(UInt32)); }
  double average() const;
  double median() const;
  double minTime() const;
  double maxTime() const;
  UInt32 maxBucket() const;
};


// Compare two histogram bucket entries by bucket value  
struct BucketCmp
{
  bool operator()(pair<HTime, UInt32> p1, pair<HTime, UInt32> p2) const
  { return p1.second < p2.second; }
};

  
// Receive states of a connection
enum RState
{
  rcv_idle,
  rcv_header,
  rcv_data
};
  

// Message type codes
enum MType
{
  mtUnknown,            // Unknown
  mtReply,              // Reply message
  mtReplyErr,           // Error reply message
  mtVersion,            // Query version
  mtWrite,              // Accept buffer of test data
  mtRead,               // Send back a buffer of test data
  mtNwrite,             // NSD style write: target asks for test data
  mtGetdata,            // Fetch data in response to NSD write request
  mtKill,               // Tell server to exit
  mtConnect,            // Connect to specified servers
  mtReset,              // Close any existing server connections
  mtRdmaConn,           // Set up RDMA connection
  mtRdmaDisconn,        // Tear down RDMA connection
  mtRdmaWrite,          // Accept data that has been written through RDMA
  mtParms,              // Set test parameters
  mtTest,               // Run performance test to all servers
  mtStatus,             // Return node status
  mtStatOn,             // Turn on test statistics gathering
  mtStatOff,            // Turn off test statistics gathering
  mtIdlePct,            // Get idle CPU percentage from last test
  mtLast                // Highest message type number
};
  

// An entry in the queue of waiters for exclusive use of a socket for sending
struct InuseWaiter
{
  // Wake this when the socket is available
  pthread_cond_t iwCond;

  // The following fields are used to determine priority between waiters
  MType mt;             // Original msgType (not mtReply or mtReplyErr)
  MsgId msgId;          // Message identifier (earlier messages are lower)
  unsigned int datalen; // Length of the message

  InuseWaiter();        // Default constructor - not defined
  InuseWaiter(MType tmt, MsgId tmsgId, unsigned int tdatalen);
};

  
// Function object to compare two InuseWaiter entries to determine which
// has higher priority
struct InuseCmp
{ bool operator()(const InuseWaiter *w1P, const InuseWaiter *w2P) const; };


// TCP connection
class TcpConn
{
  pthread_mutex_t connMutex;
  pthread_cond_t connCond;
  priority_queue<InuseWaiter *, vector<InuseWaiter *>, InuseCmp> waiters;
  map<MsgId, MsgRecord *> pendReplyTab;

  int refCount;
  Sock tcSock;
  IpAddr dest;
  MsgId nextMsgId;      // For assigning unique message identifiers
  bool inuse;           // Serializer for senders
  bool broken;          // True if socket was shut down due to error
  int cnum;             // TCP connection number
  RdmaConn *rconnP;     // RDMA connection
  list<RdmaAddr>remoteBuffs; // RDMA buffers on remote side for write tests
  list<char *>givenBuffs;    // RDMA buffers that we have given out
  Histogram connHist;   // Histogram of response times from last test
  static int nextCnum;  // Next connection number (protected by globalMutex)

  // Receiver thread state
  RState recvState;
  char *recvP;          // Receive pointer
  int recvlen;          // Length of data received so far
  int recvmax;          // Size of receive buffer
  RcvMsg *rmsgP;        // Message will be received here

public:
  TcpConn();            // Default constructor - not defined
  TcpConn(Sock tsock, IpAddr tdest);
  string destName() const { return dest.tostring(); }
  Histogram *getHistP() { return &connHist; }
  bool isBroken() const { return broken; }
  bool isRdma() const   { return rconnP != NULL; }
  Sock getSock() const  { return tcSock; }
  void holdConn();
  void releaseConn();
  void connShutdown();
  Errno receiverEvent();
  void gotMsg(RcvMsg *rmsgP);
  void getSourceAddr(MsgId msgId, char **srcAddrPP, unsigned int *srcLenP);
  Errno recvMessage();
  Errno sendMessage(MType mt, DataBuff *dbP, MsgRecord *mrP,
                    PollWait *pwaitP);
  Errno sendit(MType mt, MType origmt, MsgId msgId, DataBuff *mdbP,
               MsgRecord *mrP, PollWait *pwaitP);
  string rdmaConnect(RcvMsg *rmsgP);
  void rdmaDisconnect();
  void rdmaWrite(DataBuff *testBuffP, RdmaAddr raddr, UInt32 rlen, PollWait *pwaitP);
  void rdmaRead(RdmaAddr raddr, UInt32 rlen, char *dataP, PollWait *pwaitP);
  void rdmaSend(DataBuff *dbP, PollWait *pwaitP);
  void rdmaRecv(PollWait *pwaitP, unsigned int len);
  RdmaAddr getRemoteBuff();
  void freeRemoteBuff(RdmaAddr rBuff);
};

  
// RDMA connection
class RdmaConn
{
#ifdef RDMA
  struct ibv_qp *qp;            // Queue pair
  UInt32 rqpnum;                // Remote queue pair number
  UInt32 rlid;                  // Remote location identifier
  UInt32 rkey;                  // Remote memory key
  list<PollWait *>pwList;       // Receive requests

public:
  RdmaConn() : qp(NULL), rqpnum(0), rlid(0), rkey(0) {}
#endif
public:
  void rdPrep(TcpConn *connP, UInt32 *qpnumP, UInt32 *lidP, UInt32 *rkeyP);
  void rdPostRecv(PollWait *pwaitP);
  string rdConnect(string destName, UInt32 qpnum, UInt32 lid, UInt32 key);
  void rdDisconnect();
  void rdWrite(DataBuff *testBuffP, RdmaAddr raddr, UInt32 rlen, PollWait *pwaitP);
  void rdRead(RdmaAddr raddr, UInt32 rlen, char *dataP, PollWait *pwaitP);
  void rdSend(DataBuff *dbP, PollWait *pwaitP);
  void rdRecv(PollWait *pwaitP, unsigned int len);
};

  
// Target node for test
struct Target
{
  string hostname;
  IpAddr iaddr;                 // IP address
  TcpConn *connP;               // TCP connection (null if not connected)
  bool active;                  // For detecting inactive Targets

  Target();                     // Default constructor - not defined
  Target(const Target &m);      // Copy constructor - not defined
  Target(const string thostname, const IpAddr tiaddr)
  {
    hostname = thostname;
    iaddr = tiaddr;
    connP = NULL;
    active = true;
  }
  ~Target();
  string makeConnection();
  RcvMsg *sendm(MType mt, DataBuff *dbP = NULL, PollWait *pwaitP = NULL,
                char *srcAddrP = NULL, unsigned int srcLen = 0);
  string name() const;
};


// For sorting Target objects for round-robin scheduling
struct SortedTarget
{
  int tindex;
  Target *targP;
  SortedTarget(int ndx, Target *tP) : tindex(ndx), targP(tP) {}
  static bool comp(const SortedTarget t1, const SortedTarget t2)
  {
    if (t1.tindex != t2.tindex) return t1.tindex < t2.tindex;
    return t1.targP->iaddr < t2.targP->iaddr;
  }
};

  
// Structure to hold received messages
struct RcvMsg
{
  TcpConn *connP;               // The connection that this message came from
  char hdr[MSG_HDRSIZE];        // Buffer for receiving message header
  DataBuff msgBuff;             // Buffer for message data
  MsgId msgId;                  // Message identifier
  MType msgType;                // Message type
  string errText;               // Error message, or empty if no error

  RcvMsg();                     // Default constructor - not defined
  RcvMsg(const RcvMsg &m);      // Copy constructor - not defined
  RcvMsg(TcpConn *tconnP)
  {
    connP = tconnP;
    memset(hdr, 0, MSG_HDRSIZE);
    msgId = 0;
    msgType = mtUnknown;
    connP->holdConn();
  }
  ~RcvMsg()                     { connP->releaseConn(); }
  char *msgBuffP()              { return msgBuff.getBuffP(); }
  unsigned int msgLen() const   { return msgBuff.getBufflen(); }
  void sendReply(DataBuff *dbP, string errText = "", PollWait *pwaitP = NULL);
  bool showError();
  bool startAdminReq();
  void endAdminReq();
  void dispatch(MsgWorker *mwP);

  // Message handlers
  void handleVersion();
  void handleWrite(MsgWorker *mwP);
  void handleRdmaWrite(MsgWorker *mwP);
  void handleRead(MsgWorker *mwP);
  void handleNwrite(MsgWorker *mwP);
  void handleGetdata(MsgWorker *mwP);
  void handleConnect();
  void handleReset();
  void handleRdmaConn();
  void handleRdmaDisconn();
  void handleParms();
  void handleTest();
  void handleStatus();
  void handleIdlePct();
};


// Unique key for a message
struct MsgKey
{
  int cnum;                     // TCP connection number
  MsgId m;                      // Message ID on this connection
  MsgKey(int tcnum, MsgId tm) : cnum(tcnum), m(tm) {}
  bool operator<(const MsgKey &k) const
    { if (cnum != k.cnum) return cnum < k.cnum; return m < k.m; }
};


// This object is used to wait for pending replies
class MsgRecord
{
  pthread_mutex_t waitMutex;
  pthread_cond_t waitCond;
  set<MsgKey> waitTab;          // IDs of pending messages
  list<RcvMsg *> replies;       // Reply data

public:
  char *srcAddrP;               // Data buffer for GetData requests
  unsigned int srcLen;          // Length of the above buffer

  MsgRecord(char *tsrcAddrP = NULL, unsigned int tsrcLen = 0);
  ~MsgRecord();
  void addMsg(int cnum, MsgId msgId);
  void waitForReplies();
  bool checkReplies();
  void gotReply(int cnum, RcvMsg *rmsgP);
  RcvMsg *nextReply();
};


// An object for waiting on RDMA write requests
class PollWait
{
#ifdef RDMA
  pthread_mutex_t pwMutex;
  pthread_cond_t pwCond;
  bool complete;
  enum ibv_wc_status status;

public:
  char *mbufP;                  // Registered message buffer (optional)
  TcpConn *connP;               // Connection that message came in on
                                //   (only used for receive requests)
  PollWait() { init(); }
  PollWait(TcpConn *tconnP);
  ~PollWait();
  void init();
  ibv_wc_status wait();
  void wakeup(ibv_wc_status s);
#else
public:
  PollWait() { mbufP = NULL; }
  char *mbufP;
#endif // RDMA
};

  
// Test types
enum TType
{
  ttWrite,              // Round-robin write
  ttRead,               // Round-robin read
  ttNwrite,             // NSD-style write
  ttRW,                 // Simultaneous read and write
  ttSwrite,             // Single node write per tester
  ttSread,              // Single node read per tester
  ttLast
};


// An object to hold info about a node used by the tester thread
struct TNodeInfo
{
  Target *targP;        // Target node
  RdmaAddr tBuff;       // RDMA memory buffer on target for write tests
  TNodeInfo(Target *tP, RdmaAddr tb) : targP(tP), tBuff(tb) {}
};


// Tester thread work request
struct TestReq
{
  TType tt;
  list<TNodeInfo> testNodes;
  HTime endTime;
  UInt64 totBytes;
  Histogram hist;
  string errText;

  TestReq(TType ttt, HTime tendTime)
  {
    tt = ttt;
    endTime = tendTime;
    totBytes = 0;
  }
};

  
// An entry in the newSockets list of the receiver thread
struct NewSock
{
  Sock sock;
  TcpConn *connP;
  NewSock(Sock tsock, TcpConn *tconnP) : sock(tsock), connP(tconnP) {}
};

  
// A receiver thread
class Receiver : public Thread
{
  pthread_mutex_t receiverMutex;        // Protects newSockets
  pthread_cond_t receiverCond;          // For additions to newSockets
  list<NewSock> newSockets;             // New sockets for receiver thread
  Sock rcvSocks[2];                     // Socket pair for waking up poll

public:
  Receiver();
  virtual ~Receiver();
  void addConn(Sock sock, TcpConn *connP);
  void wakeUp();
  virtual int threadBody();
};


// Message worker thread
class MsgWorker : public Thread
{
public:
  char *rdBuffP;                // A memory buffer to use for RDMA requests
  unsigned int rdLen;           // Length of the above buffer
  PollWait pwait;               // For waiting on RDMA requests
  DataBuff rtestBuff;           // Test data for read replies

  MsgWorker() : rdBuffP(NULL), rdLen(0) {}
  virtual ~MsgWorker() {}
  void getRdmaBuff();
  void freeRdmaBuff();
  virtual int threadBody();
};


// Tester thread
class Tester : public Thread
{
  TestReq *reqP;
  
public:
  Tester() : reqP(NULL) {}
  virtual ~Tester() {}
  virtual int threadBody();
  void doTest(TestReq *trP);
};


// Listen/accept thread
class ListenAccept : public Thread
{
  int currsize;                 // Socket buffer size currently in use
public:
  ListenAccept() : currsize(0) {}
  virtual ~ListenAccept() {}
  virtual int threadBody();
  void updateSocksize();
};


// RDMA receiver thread
class RdmaReceiver : public Thread
{
public:
  RdmaReceiver() {}
  virtual ~RdmaReceiver() {}
  virtual int threadBody();
};


// RDMA async event handler thread
class RdmaAsync : public Thread
{
public:
  RdmaAsync() {}
  virtual ~RdmaAsync() {}
  virtual int threadBody();
};


// Global variables
static string progname;
static bool littleEndian;
static bool quitflag = false;
static unsigned long pagesize;
static deque<string> commands;
static int nestingLevel = 0;
static int idleTime = -1;
static bool collectStats = false;
static unsigned int mbufSize = 0;

static multimap<IpAddr, Target *> serverNodes;
static map<IpAddr, Target *> clientNodes;
static map<IpAddr, Target *> allNodes;

static int debugLevel = 0;
static short port = NSDPERF_PORT;
static string rdmaDev;
static unsigned char rdmaPort = 1;
static int socksize = 0;
static int nWorkers = MSG_WORKERS;
static int nParallel = 1;
static enum RdmaMode { rOff, rOn, rAll, rInline } useRdma = rOff;
static bool rdmaInitialized = false;
static bool server = false;
static bool useipv6 = false;
static unsigned int nClients = 0;
static int addrFamily;
static int testTime = 10;        
static unsigned int buffsize = 4194304;
static unsigned int nTesterThreads = TESTER_THREADS;
static bool showHist = false;
static bool verify = false;
static string plotFname;

static pthread_mutex_t logMutex;
static pthread_mutex_t globalMutex;
static pthread_cond_t globalCond;
static list<Thread *> deadThreads;
static unsigned int nThreadsStarted = 0;
static vector<Receiver *> receiverTab;
static vector<Receiver *>::iterator nextReceiver;
static bool cmdInProgress = false;
static ListenAccept *laThreadP = NULL;
int TcpConn::nextCnum = 0;

static pthread_mutex_t workerMutex;
static pthread_cond_t workerCond;
static list<RcvMsg *> msgQueue;
static list<RcvMsg *> bulkQueue;
static vector<MsgWorker *> workerTab;
static int workersActive = 0;

static pthread_mutex_t testerMutex;
static pthread_cond_t testerCond;
static set<Tester *> testerTab;
static list<TestReq *> doneList;
static unsigned int nTestersWorking = 0;



// Convert seconds to HTime
static HTime sectoht(int sec)
{ return sec * 1000000000LL; }


// Convert HTime to seconds
static double httosec(HTime t)
{ return t / 1000000000.0; }


// Convert HTime to milliseconds
static long long httomsec(HTime t)
{ return (t + 500000) / 1000000; }


// Return current time of day
static HTime getTime()
{
#ifdef _POSIX_TIMERS
  struct timespec ts;
  long rc = clock_gettime(CLOCK_REALTIME, &ts);
  if (rc != 0) Error("clock_gettime rc " << rc);
  return sectoht(ts.tv_sec) + ts.tv_nsec;
#else
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return sectoht(tv.tv_sec) + tv.tv_usec * 1000;
#endif
}


// Sleep until specified time.  Return current time.
static HTime sleepUntil(HTime endTime)
{
  HTime now;
  for (now = getTime(); now < endTime; now = getTime())
    usleep((endTime - now) / 1000);
  return now;
}


// Swap bytes in a word
static UInt32 ByteSwap32(UInt32 d)
{
  return ((d >> 24) & 0x000000FFU) |
         ((d >>  8) & 0x0000FF00U) |
         ((d <<  8) & 0x00FF0000U) |
         ((d << 24) & 0xFF000000U);
}


// Swap bytes in a double word
static UInt64 ByteSwap64(UInt64 d)
{
  return ((d >> 56) & 0x00000000000000FFULL) |
         ((d >> 40) & 0x000000000000FF00ULL) |
         ((d >> 24) & 0x0000000000FF0000ULL) |
         ((d >>  8) & 0x00000000FF000000ULL) |
         ((d <<  8) & 0x000000FF00000000ULL) |
         ((d << 24) & 0x0000FF0000000000ULL) |
         ((d << 40) & 0x00FF000000000000ULL) |
         ((d << 56) & 0xFF00000000000000ULL);
}


// Report the number of online processors
static int numProcessors()
{
  #ifndef _SC_NPROCESSORS_ONLN
    return 2;
  #else
    return sysconf(_SC_NPROCESSORS_ONLN);
  #endif
}


// Set socket send/receive buffer sizes.  This routine must be called
// before listen or connect.
static void setSockSizes(Sock sock)
{
  if (socksize == 0)
    return;
  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &socksize, sizeof(socksize)) < 0)
    Warnm("setsockopt SO_RCVBUF failed");
  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &socksize, sizeof(socksize)) < 0)
    Warnm("setsockopt SO_SNDBUF failed");
}


// Make a socket non-blocking
static void setSockNonblocking(Sock sock)
{
  int rc = ioctl(sock, (int)FIONBIO, (char *)&on, (int)sizeof(int));
  if (rc < 0)
    Errorm("ioctl FIONBIO");
}


// Get error message for a given error number
static string geterr(int e)
{
  char b[256];
  snprintf(b, sizeof(b), "error %d", e);
  (void)strerror_r(e, b, sizeof(b));
  return b;
}


// Get next word from string starting at pos.  Update pos to end of word.
static string getword(const string s, string::size_type &pos,
                      const char *delim = " \t\r")
{
  string::size_type pos1;

  pos1 = s.find_first_not_of(delim, pos);
  if (pos1 == string::npos)
  {
    pos = pos1;
    return "";
  }
  pos = s.find_first_of(delim, pos1);
  return s.substr(pos1, (pos == string::npos) ? pos : pos - pos1);
}


// Return true if start of line matches specified string
static bool match(const string line, const string s)
{
  return line.length() >= s.length() && line.substr(0, s.length()) == s;
}


// Split input string into separate command lines, delimited by semicolon
// or slash+space.
static void splitCmd(const string inp, deque<string> *linesP)
{
  int p1, p2, len = inp.length();
  for (p1 = 0; p1 < len; p1 = p2 + 1)
  {
    for (p2 = p1; p2 < len; p2++)
      if (inp[p2] == ';' ||
          (inp[p2] == '/' && p2 + 1 < len && inp[p2 + 1] == ' '))
        break;
    linesP->push_back(inp.substr(p1, p2 - p1));
    if (p2 < len && inp[p2] == '/')
      p2++;
  }
}


// Round d to the specified number of significant decimal digits.
// Negative values are not supported.
static double siground(double d, int digits)
{
  int j, n;
  double tpow = 1.0;

  if (d <= 0.0 || digits <= 0) return 0.0;
  for (j = 0; j < digits; j++) tpow *= 10.0;

  // Normalize
  n = 0;
  if (d >= tpow) while (d >= tpow)        d /= 10.0, n--;
  else           while (d * 10.0 < tpow)  d *= 10.0, n++;

  // Round to precision
  d = floor(d + 0.5);

  // Undo normalize
  if (n > 0)      for (j = 0; j < n; j++)  d /= 10.0;
  else if (n < 0) for (j = 0; j < -n; j++) d *= 10.0;

  return d;
}


// Allocate a new data buffer for output.  If a big enough buffer is
// already allocated, just adjust the assigned buffer length.
void DataBuff::newBuff(unsigned int len)
{
  if (len == 0)
    Error("Allocating zero length buffer");

  bufflen = len;
  buffpos = 0;
  if (alloc > 0)
  {
    if (len <= alloc)
      return;
    delete [] buffP;
  }
  buffP = new char[len];
  alloc = len;
}


// Initialize buffer to point to specified data area
void DataBuff::initBuff(char *dataP, unsigned int datalen)
{
  if (alloc > 0)
  {
    delete [] buffP;
    alloc = 0;
  }
  buffP = dataP;
  bufflen = datalen;
  buffpos  = 0;
}


// Fill buffer with random data
void DataBuff::fillBuff(UInt64 seed)
{
  char *p;
  for (p = buffP; p - buffP + sizeof(UInt64) <= bufflen; p += sizeof(UInt64))
  {
    UInt64 *p64 = reinterpret_cast<UInt64 *>(p);
    *p64 = littleEndian ? seed : ByteSwap64(seed);
    seed += SCRAMBLE;
  }
  while (p < buffP + bufflen)
    *p++ = '\0';
}


// Verify buffer contents.  Return true if contents are correct.
bool DataBuff::verifyBuff(UInt64 seed)
{
  char *p;
  for (p = buffP; p - buffP + sizeof(UInt64) <= bufflen; p += sizeof(UInt64))
  {
    UInt64 *p64 = reinterpret_cast<UInt64 *>(p);
    if (*p64 != (littleEndian ? seed : ByteSwap64(seed)))
      return false;
    seed += SCRAMBLE;
  }
  while (p < buffP + bufflen)
    if (*p++ != '\0')
      return false;
  return true;
}


// Put a UInt32 item into buffer, performing endian conversion if
// necessary, and bump the buffer pointer.
void DataBuff::putUInt32(UInt32 i)
{
  if (buffpos + sizeof(i) > bufflen) Error("putUInt32 buffer overflow");
  if (!littleEndian) i = ByteSwap32(i);
  memcpy(&buffP[buffpos], &i, sizeof(i));
  buffpos += sizeof(i);
}


// Put a UInt64 item into buffer, performing endian conversion if
// necessary, and bump the buffer pointer.
void DataBuff::putUInt64(UInt64 i)
{
  if (buffpos + sizeof(i) > bufflen) Error("putUInt64 buffer overflow");
  if (!littleEndian) i = ByteSwap64(i);
  memcpy(&buffP[buffpos], &i, sizeof(i));
  buffpos += sizeof(i);
}


// Put an RdmaAddr item into buffer, performing endian conversion if
// necessary, and bump the buffer pointer.
void DataBuff::putRdmaAddr(RdmaAddr a)
{
  putUInt64(a);
}


// Put an HTime item into buffer, performing endian conversion if
// necessary, and bump the buffer pointer.
void DataBuff::putHTime(HTime t)
{
  union { UInt64 i; HTime t; } u;
  u.t = t;
  if (buffpos + sizeof(u.i) > bufflen) Error("putHTime buffer overflow");
  if (!littleEndian) u.i = ByteSwap64(u.i);
  memcpy(&buffP[buffpos], &u.i, sizeof(u.i));
  buffpos += sizeof(u.i);
}


// Put an IpAddr into buffer and bump the buffer pointer.  No endian
// conversion is performed on the address data, since it is kept in memory
// in native byte order.
void DataBuff::putIpAddr(IpAddr iaddr)
{
  if (buffpos + sizeof(UInt32) + sizeof(iaddr.a) > bufflen)
    Error("putIpAddr buffer overflow");
  putUInt32(iaddr.fam);
  memcpy(&buffP[buffpos], iaddr.a, sizeof(iaddr.a));
  buffpos += sizeof(iaddr.a);
}


// Put string into buffer, with length and padding
void DataBuff::putString(string s)
{
  unsigned int slen = s.length();
  unsigned int padlen = 4 - (slen & 3);
  putUInt32(slen + padlen);
  if (buffpos + slen + padlen > bufflen) Error("putString buffer overflow");
  s.copy(&buffP[buffpos], slen);
  for (buffpos += slen; padlen > 0; padlen--)
    buffP[buffpos++] = '\0';
}


// Get UInt32 from buffer
UInt32 DataBuff::getUInt32()
{
  UInt32 i;
  if (buffpos + sizeof(i) > bufflen) Error("getUInt32 buffer underflow");
  memcpy(&i, &buffP[buffpos], sizeof(i));
  buffpos += sizeof(i);
  return littleEndian ? i : ByteSwap32(i);
}


// Get UInt64 from buffer
UInt64 DataBuff::getUInt64()
{
  UInt64 i;
  if (buffpos + sizeof(i) > bufflen) Error("getUInt64 buffer underflow");
  memcpy(&i, &buffP[buffpos], sizeof(i));
  buffpos += sizeof(i);
  return littleEndian ? i : ByteSwap64(i);
}


// Get RdmaAddr from buffer
RdmaAddr DataBuff::getRdmaAddr()
{
  return getUInt64();
}


// Get HTime from buffer
HTime DataBuff::getHTime()
{
  union { UInt64 i; HTime t; } u;
  if (buffpos + sizeof(u.i) > bufflen) Error("getHTime buffer underflow");
  memcpy(&u.i, &buffP[buffpos], sizeof(u.i));
  buffpos += sizeof(u.i);
  if (!littleEndian) u.i = ByteSwap64(u.i);
  return u.t;
}


// Get IpAddr from buffer
IpAddr DataBuff::getIpAddr()
{
  IpAddr iaddr;
  if (buffpos + sizeof(UInt32) + sizeof(iaddr.a) > bufflen)
    Error("getIpAddr buffer underflow");
  iaddr.fam = static_cast<Atype>(getUInt32());
  memcpy(iaddr.a, &buffP[buffpos], sizeof(iaddr.a));
  buffpos += sizeof(iaddr.a);
  return iaddr;
}


// Get string from buffer
string DataBuff::getString()
{
  unsigned int slen = getUInt32();
  if (buffpos + slen > bufflen) Error("getString buffer underflow");
  string s(&buffP[buffpos]);
  if (s.length() >= slen) Error("incorrect getString length");
  buffpos += slen;
  return s;
}


// Calculate buffer space needed to do putString
static unsigned int calcLen(string s)
{
  unsigned int slen = s.length();
  unsigned int padlen = 4 - (slen & 3);
  return sizeof(UInt32) + slen + padlen;
}


// Put Histogram into buffer and bump the buffer pointer
void Histogram::putBuff(DataBuff *dbP) const
{
  map<HTime, UInt32>::const_iterator b;
  dbP->putUInt64(totalTime);
  dbP->putUInt32(nEvents);
  dbP->putUInt32(buckets.size());
  for (b = buckets.begin(); b != buckets.end(); ++b)
  {
    dbP->putHTime(b->first);
    dbP->putUInt32(b->second);
  }
}


// Get Histogram from buffer
void Histogram::getBuff(DataBuff *dbP)
{
  unsigned int nb;
  HTime bnum;
  UInt32 count;
  buckets.clear();
  totalTime = dbP->getUInt64();
  nEvents = dbP->getUInt32();
  for (nb = dbP->getUInt32(); nb > 0; nb--)
  {
    bnum = dbP->getHTime();
    count = dbP->getUInt32();
    buckets[bnum] = count;
  }
}


// Structure to hold CPU utilization statistics in ticks
struct CpuStats
{
  UInt64 user, system, idle, iowait, other;
  UInt64 total() { return user + system + idle + iowait + other; }
};

  
// Get CPU utilization statistics.  Return false if not supported.
static bool getStats(CpuStats *statsP)
{
  static bool supported = true;
  if (!supported)
    return false;

#ifdef _AIX
  perfstat_partition_total_t pstats;
  if (perfstat_partition_total(NULL, &pstats, sizeof(pstats), 1) < 0)
    Errorm("perfstat_partition_total");

  // More work is required to calculate utilization for LPAR in shared mode,
  // so punt for now.
  if (pstats.type.b.shared_enabled)
  {
    supported = false;
    return false;
  }

  statsP->user   = pstats.puser;
  statsP->system = pstats.psys;
  statsP->idle   = pstats.pidle;
  statsP->iowait = pstats.pwait;
  statsP->other  = 0;
  return true;

#else
  ifstream infile;
  string line;
  UInt64 n[8];
  bool gotit = false;

  // Get CPU info from /proc/stat.  See Documentation/filesystems/proc.txt
  // in the Linux kernel source.
  infile.open("/proc/stat");
  if (!infile)
  {
    supported = false;
    return false;
  }
  while (getline(infile, line))
  {
    // Fields are: user nice system idle iowait irq softirq steal
    if (sscanf(line.c_str(), "cpu %llu %llu %llu %llu %llu %llu %llu %llu",
               &n[0], &n[1], &n[2], &n[3], &n[4], &n[5], &n[6], &n[7]) == 8)
    {
      gotit = true;
      break;
    }
    
    // Older kernels didn't have the steal field
    n[7] = 0;
    if (sscanf(line.c_str(), "cpu %llu %llu %llu %llu %llu %llu %llu",
               &n[0], &n[1], &n[2], &n[3], &n[4], &n[5], &n[6]) == 7)
    {
      gotit = true;
      break;
    }
  }
  infile.close();
  if (!gotit)
  {
    supported = false;
    return false;
  }

  statsP->user   = n[0] + n[1];
  statsP->system = n[2];
  statsP->idle   = n[3];
  statsP->iowait = n[4];
  statsP->other  = n[5] + n[6] + n[7];
  return true;
#endif
}


// Return CPU idle percentage since previous call, or -1 if not supported
// on this OS or if this is the first call.
static int cpuIdle()
{
  static bool initialized = false;
  static CpuStats oldstats;
  CpuStats stats;
  UInt64 tot;
  int val = -1;

  if (!getStats(&stats))
    return -1;

  if (initialized)
  {
    tot = stats.total() - oldstats.total();
    if (tot != 0)
      val = (stats.idle - oldstats.idle) * 100 / tot;
  }
  else
    initialized = true;

  oldstats = stats;
  return val;
}


// Convert hostname and IP address to a string for output messages
static string hostString(const string hname, const IpAddr iaddr)
{
  int rc;
  struct sockaddr_storage saddr;

  // Determine whether the host name is a dotted decimal or hex IP address
  // by trying to convert it to a network address structure.  If the
  // conversion is successful, then don't show the IP address in result,
  // since that would be redundant.
  rc = inet_pton(iaddr.getFamily(), hname.c_str(), &saddr);
  if (rc < 0)
    Errorm("inet_pton");
  else if (rc == 0)
    return hname + " (" + iaddr.tostring() + ")";
  return hname;
}


// Error-checking wrappers for mutex and condition variable routines
static void thInitMutex(pthread_mutex_t *mutexP)
{ if (pthread_mutex_init(mutexP, NULL) != 0) Errorm("pthread_mutex_init"); }

static void thInitCond(pthread_cond_t *condP)
{ if (pthread_cond_init(condP, NULL) != 0) Errorm("pthread_cond_init"); }

static void thLock(pthread_mutex_t *mutexP)
{ if (pthread_mutex_lock(mutexP) != 0) Errorm("pthread_mutex_lock"); }

static void thUnlock(pthread_mutex_t *mutexP)
{ if (pthread_mutex_unlock(mutexP) != 0) Errorm("pthread_mutex_unlock"); }

static void thWait(pthread_cond_t *condP, pthread_mutex_t *mutexP)
{ if (pthread_cond_wait(condP, mutexP) != 0) Errorm("pthread_cond_wait"); }

static void thSignal(pthread_cond_t *condP)
{ if (pthread_cond_signal(condP) != 0) Errorm("pthread_cond_signal"); }

static void thBcast(pthread_cond_t *condP)
{ if (pthread_cond_broadcast(condP) != 0) Errorm("pthread_cond_broadcast"); }

static void thJoin(pthread_t th, void **valPP)
{ if (pthread_join(th, valPP) != 0) Errorm("pthread_join"); }

static void thKill(pthread_t th, int sig)
{ if (pthread_kill(th, sig) != 0) Errorm("pthread_kill"); }


// Return a random 64-bit number
static UInt64 randSeed()
{
  thLock(&globalMutex);
  UInt64 w = random();
  w = (w << 32) + random();
  thUnlock(&globalMutex);
  return w;
}


// IpAddr constructor
IpAddr::IpAddr()
{
  in_addr_t addr = INADDR_NONE;
  fam = AT_IPV4;
  memset(a, 0, sizeof(a));
  memcpy(a, &addr, sizeof(addr));
}


// Compare two IP addresses
bool IpAddr::operator<(const IpAddr &iaddr) const
{
  unsigned int j;
  if (fam != iaddr.fam)
    return fam < iaddr.fam;
  for (j = 0; j < sizeof(a); j++)
    if (a[j] != iaddr.a[j])
      return a[j] < iaddr.a[j];
  return false;
}


// Test inequality of two IP addresses
bool IpAddr::operator!=(const IpAddr &iaddr) const
{
  unsigned int j;
  if (fam != iaddr.fam)
    return true;
  for (j = 0; j < sizeof(a); j++)
    if (a[j] != iaddr.a[j])
      return true;
  return false;
}


// Convert a host name string to an IP address.  If an error occurs, log an
// error message and return E_INVAL.
Errno IpAddr::parse(const string hostname)
{
  string::size_type pos = 0;
  string hname = getword(hostname, pos);
#ifdef IPV6_SUPPORT
  struct addrinfo hints, *resP;
  int rc;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = addrFamily;
  hints.ai_socktype = SOCK_STREAM;

  rc = getaddrinfo(hname.c_str(), NULL, &hints, &resP);
  if (rc != 0)
  {
    Log("Invalid hostname \"" << hname << "\": " << gai_strerror(rc));
    return E_INVAL;
  }

  // Copy the address data.  If multiple addresses are returned, we only
  // use the first one.  It would be better to try all of them until a
  // working address is found.
  loadSockaddr(resP->ai_addr);

  freeaddrinfo(resP);
#else
  unsigned int d[4];
  fam = AT_IPV4;
  memset(a, 0, sizeof(a));
  if (sscanf(hname.c_str(), "%u.%u.%u.%u", &d[0], &d[1], &d[2], &d[3]) == 4)
  {
    if (d[0] > 255 || d[1] > 255 || d[2] > 255 || d[3] > 255)
    {
      Log("Invalid hostname: " << hname);
      return E_INVAL;
    }
    a[0] = d[0]; a[1] = d[1]; a[2] = d[2]; a[3] = d[3];
  }
  else
  {
    struct hostent *hp = gethostbyname(hname.c_str());
    if (hp == NULL || hp->h_addrtype != AF_INET || hp->h_length != 4)
    {
      Log("Cannot find " << hname);
      return E_INVAL;
    }
    memcpy(a, hp->h_addr, 4);
  }
#endif
  return E_OK;
}


// Load IpAddr contents from a sockaddr struct
void IpAddr::loadSockaddr(const sockaddr *saddrP)
{
  switch (saddrP->sa_family)
  {
#ifdef IPV6_SUPPORT
    case AF_INET6:
      fam = AT_IPV6;
      memcpy(a, &reinterpret_cast<const sockaddr_in6 *>(saddrP)->sin6_addr,
             sizeof(a));
      break;
#endif

    case AF_INET:
      fam = AT_IPV4;
      memset(a, 0, sizeof(a));
      memcpy(a, &reinterpret_cast<const sockaddr_in *>(saddrP)->sin_addr,
             sizeof(in_addr_t));
      break;

    default:
      Error("Invalid sa_family");
      break;
  }
}


// Convert IP address to printable string
string IpAddr::tostring() const
{
#ifdef IPV6_SUPPORT
  char buf[INET6_ADDRSTRLEN];
#else
  char buf[INET_ADDRSTRLEN];
#endif
  if (inet_ntop(getFamily(), &a, buf, sizeof(buf)) == NULL)
    Errorm("inet_ntop");
  return buf;  
}


#ifdef RDMA
// RDMA data
static struct ibv_context *ibContext = NULL; // Device context
static struct ibv_device_attr ibAttr;
static struct ibv_comp_channel *ibCC;
static struct ibv_pd *ibPD;
static struct ibv_cq *ibCQ;
static struct ibv_mr *ibMR;
static int cqSize = 0;                  // Number of entries in CQ
static RdmaReceiver *rdmaRcvThreadP = NULL;
static RdmaAsync *rdmaAsyncThreadP = NULL;
static bool rdmaRcvRun = false;         // Set to false to kill receive thread
static bool rdmaAsyncRun = false;       // Set to false to kill async thread
static char *memoryPoolP = NULL;        // Registered memory pool
static RdmaAddr memoryPoolBase;         // Base address for buffers in pool
static unsigned int poolCount;          // Count of buffers in the pool
static unsigned int poolBuffsize;       // Size of a buffer in the pool
static list<char *> poolList;           // List of free memory buffers
static unsigned int mbufCount;          // Count of message buffers in the pool
static unsigned int mbufBuffsize;       // Size of a message buffer in the pool
static list<char *> mbufList;           // List of free message buffers


// Free the RDMA memory pool if it was allocated
static void rdmaMemoryFree()
{
  if (memoryPoolP == NULL)
    return;
  if (poolList.size() != poolCount || mbufList.size() != mbufCount)
    Error("RDMA memory pool freed while buffers in use");
  if (ibv_dereg_mr(ibMR) != 0) Errorm("ibv_dereg_mr failed");
  delete [] memoryPoolP;
  memoryPoolP = NULL;
  poolList.clear();
  mbufList.clear();
}


// Allocate a memory pool and register it with the RDMA device.  If a pool
// already exists, re-use it if the test parameters haven't changed.
// Otherwise destroy old pool and re-create.  The protection domain (ibPD)
// must have already been allocated.
static void rdmaMemoryAlloc()
{
  Int64 poolLen;
  unsigned int buffsizeRounded, mbufSizeRounded, nConn,
    newPoolCount, newMbufCount;

  // Buffer read or write ops are initiated by tester threads on the
  // clients, so the total number of outstanding requests can never be more
  // than the number of testers times the number of clients.  Thus, if we
  // allocate space for that many buffers on each connection, we should
  // always have room to do an operation without having to wait for a
  // buffer.
  //
  // When a server handles an NSD write request, the message worker thread
  // needs a buffer to use for the target of a Getdata RPC.  Rather than
  // allocating these on the fly, which could cause contention among the
  // worker threads, we use a dedicated buffer for each worker.  
  //
  // Theoretically, the tester threads could share buffers with the worker
  // threads, since a node is either a client or a server, and thus only
  // testers or workers will be using the buffers during any particular
  // test run.  However, it is easier for worker threads to always allocate
  // buffers up front rather than trying to give them back for use by
  // tester threads when needed.  Therefore, we reserve buffer space for
  // each worker thread in addition to the space used by tester threads.
  //
  // Add one more buffer per tester thread to use on the sending side to
  // hold test data, and another per worker to reply to read requests.
  nConn = nClients * nParallel;
  newPoolCount = nTesterThreads * nConn + nWorkers * 2 + nTesterThreads;

  // Message buffers in registered memory (mbufs) are used for sending and
  // receiving when RDMA is set to "all".  These are allocated up front:
  //   1) Each worker thread gets a PollWait object with mbuf
  //   2) Each tester thread gets an mbuf for sending
  //   3) On servers, incoming connections get an mbuf for each tester
  //   4) On clients, outgoing connections get an mbuf for each tester
  if (serverNodes.size() > nConn)
    nConn = serverNodes.size();
  if (useRdma == rAll || useRdma == rInline)
    newMbufCount = nWorkers + nTesterThreads + nConn * nTesterThreads;
  else
    newMbufCount = 0;

  // When sending data buffers inline, an mbuf must be big enough to read
  // the whole thing.
  mbufSize = MSG_HDRSIZE + MAX_RPCSIZE;
  if (useRdma == rInline)
    mbufSize += buffsize;

  buffsizeRounded = (buffsize + pagesize - 1) / pagesize * pagesize;
  mbufSizeRounded = (mbufSize + pagesize - 1) / pagesize * pagesize;

  if (memoryPoolP != NULL)
  {
    if (poolList.size() != poolCount || mbufList.size() != mbufCount)
      Error("RDMA memory pool allocated while buffers in use");

    // If we already have a pool of the correct size, then use it
    if (buffsizeRounded == poolBuffsize && poolCount == newPoolCount &&
        mbufSizeRounded == mbufBuffsize && mbufCount == newMbufCount)
      return;

    rdmaMemoryFree();
  }

  // Create a memory pool of page-aligned buffers.  These don't really have
  // to be page-aligned, but it might reduce contention between threads
  // (although the chances that this will make any difference are slim).
  poolCount = newPoolCount;
  poolBuffsize = buffsizeRounded;
  mbufCount = newMbufCount;
  mbufBuffsize = mbufSizeRounded;
  poolLen = static_cast<Int64>(poolBuffsize) * poolCount +
            static_cast<Int64>(mbufBuffsize) * mbufCount;
  memoryPoolP = new char[poolLen + pagesize];
  char *poolPosP = memoryPoolP +
    (pagesize - reinterpret_cast<unsigned long>(memoryPoolP) % pagesize);
  memoryPoolBase = poolPosP;

  // Register the entire pool
  int accFlags = IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ |
    IBV_ACCESS_REMOTE_WRITE;
  ibMR = ibv_reg_mr(ibPD, poolPosP, poolLen, accFlags);
  if (ibMR == NULL) Errorm("ibv_reg_mr failed");

  // Initialize free space list
  for (; newPoolCount > 0; newPoolCount--)
  {
    poolList.push_back(poolPosP);
    poolPosP += poolBuffsize;
  }
  for (; newMbufCount > 0; newMbufCount--)
  {
    mbufList.push_back(poolPosP);
    poolPosP += mbufBuffsize;
  }
}


// Get some memory for a data buffer.  If RDMA is initialized, get it from
// from the registered RDMA memory pool.  Otherwise get from heap.
static char *poolGet(unsigned int len)
{
  char *buffP;
  if (!rdmaInitialized)
    return new char[len];
  if (memoryPoolP == NULL)
    Error("memory pool not initialized");
  if (len > poolBuffsize)
    Error("incorrect length in memory pool request");
  thLock(&globalMutex);
  if (poolList.empty())
    Error("memory pool exhausted");
  buffP = poolList.front();
  poolList.pop_front();
  thUnlock(&globalMutex);
  return buffP;
}


// Give back some memory that was allocated with poolGet.
static void poolFree(char *buffP)
{
  if (buffP == NULL)
    return;
  if (!rdmaInitialized)
  {
    delete [] buffP;
    return;
  }
  if (memoryPoolP == NULL)
    Error("memory pool not initialized");
  thLock(&globalMutex);
  poolList.push_back(buffP);
  thUnlock(&globalMutex);
}


// Get some memory for a message buffer for RDMA send and receive requests.
static char *mbufGet(unsigned int len)
{
  char *mbufP;
  if (!rdmaInitialized)
    Error("message buffer allocated without initializing RDMA");
  if (memoryPoolP == NULL)
    Error("memory pool not initialized");
  if (len > mbufBuffsize)
    Error("incorrect length in message buffer request");
  thLock(&globalMutex);
  if (mbufList.empty())
    Error("message buffer pool exhausted");
  mbufP = mbufList.front();
  mbufList.pop_front();
  thUnlock(&globalMutex);
  return mbufP;
}


// Give back some memory that was allocated with mbufGet.
static void mbufFree(char *mbufP)
{
  if (mbufP == NULL)
    return;
  if (!rdmaInitialized)
    Error("message buffer freed without initializing RDMA");
  if (memoryPoolP == NULL)
    Error("memory pool not initialized");
  thLock(&globalMutex);
  mbufList.push_back(mbufP);
  thUnlock(&globalMutex);
}


// Open RDMA device and initialize it if necessary.  This must be called
// after test parameters have been established, since it uses those
// parameters to adjust queue sizes and memory allocation.  Returns error
// message if failure.
static string rdmaStart()
{
  if (nClients == 0)
    Error("RDMA start with no clients");

  // Set memory locking limits to the maximum allowed
  struct rlimit rlim;
  if (getrlimit(RLIMIT_MEMLOCK, &rlim) != 0)
    Errorm("getrlimit RLIMIT_MEMLOCK failed");
  rlim.rlim_cur = rlim.rlim_max;
  if (setrlimit(RLIMIT_MEMLOCK, &rlim) != 0)
    Errorm("setrlimit RLIMIT_MEMLOCK to" << rlim.rlim_cur << " failed");

  if (ibContext == NULL)
  {
    struct ibv_device *devP, **devList;
    int j, numDevices;

    devList = ibv_get_device_list(&numDevices);
    if (devList == NULL) Errorm("ibv_get_device_list");
    if (numDevices < 1)
    {
      ibv_free_device_list(devList);
      return "No RDMA devices found";
    }

    // If no device name was specified, open the first one.  Otherwise,
    // look for the requested one.
    if (rdmaDev.length() == 0)
      devP = devList[0];
    else
    {
      for (j = 0; j < numDevices; j++)
      {
        devP = devList[j];
        if (!strcmp(ibv_get_device_name(devP), rdmaDev.c_str()))
          break;
      }
      if (j == numDevices)
        Error("RDMA device " << rdmaDev << " was not found");
    }
    ibContext = ibv_open_device(devP);
    ibv_free_device_list(devList);
    if (ibContext == NULL)
      return "Open of RDMA device failed";

    // Save device attributes
    if (ibv_query_device(ibContext, &ibAttr) != 0)
      Errorm("ibv_query_device failed");

    // Create a completion channel
    ibCC = ibv_create_comp_channel(ibContext);
    if (ibCC == NULL) Error("ibv_create_comp_channel failed");

    // Allocate protection domain
    ibPD = ibv_alloc_pd(ibContext);
    if (ibPD == NULL) Error("ibv_alloc_pd failed");

    rdmaInitialized = true;
  }
  
  // If we have an existing completion queue with the correct size, use it.
  // If it has the wrong size, kill the RdmaReceiver thread (which waits
  // for events on the queue), and destroy the queue.  Then create a new
  // queue and start the RdmaReceiver thread..
  int newcqSize = nTesterThreads * nClients + nWorkers;
  if (newcqSize > ibAttr.max_cqe)
    newcqSize = ibAttr.max_cqe;
  if (cqSize != newcqSize)
  {
    if (cqSize > 0)
    {
      // Kill RDMA receiver thread and wait for it to go away
      if (rdmaRcvThreadP == NULL)
        Error("completion queue exists without RDMA receiver");
      rdmaRcvRun = false;
      thKill(rdmaRcvThreadP->getThread(), SIGUSR1);
      rdmaRcvThreadP->waitForDeath();

      // Kill RDMA async event handler thread and wait for it to go away
      if (rdmaAsyncThreadP == NULL)
        Error("completion queue exists without RDMA async handler");
      rdmaAsyncRun = false;
      thKill(rdmaAsyncThreadP->getThread(), SIGUSR1);
      rdmaAsyncThreadP->waitForDeath();

      // Destroy the old queue
      if (ibv_destroy_cq(ibCQ) != 0) Errorm("ibv_destroy_cq failed");
    }

    // Create the completion queue
    cqSize = newcqSize;
    ibCQ = ibv_create_cq(ibContext, cqSize, NULL, ibCC, 0);
    if (ibCQ == NULL) Error("ibv_create_cq failed");

    // Start RdmaReceiver thread
    rdmaRcvRun = true;
    rdmaRcvThreadP = new RdmaReceiver;
    rdmaRcvThreadP->init();

    // Start RdmaAsync thread
    rdmaAsyncRun = true;
    rdmaAsyncThreadP = new RdmaAsync;
    rdmaAsyncThreadP->init();
  }

  // Allocate memory pool, possibly destroying an old one first
  rdmaMemoryAlloc();
  return "";
}


// Shut down RDMA device and close it
static void rdmaShutdown()
{
  if (ibContext == NULL)
    return;
  
  // Kill RDMA receiver thread and wait for it to go away
  if (rdmaRcvThreadP != NULL)
  {
    rdmaRcvRun = false;
    thKill(rdmaRcvThreadP->getThread(), SIGUSR1);
    rdmaRcvThreadP->waitForDeath();
    rdmaRcvThreadP = NULL;
  }

  // Kill RDMA async event handler thread and wait for it to go away
  if (rdmaAsyncThreadP != NULL)
  {
    rdmaAsyncRun = false;
    thKill(rdmaAsyncThreadP->getThread(), SIGUSR1);
    rdmaAsyncThreadP->waitForDeath();
    rdmaAsyncThreadP = NULL;
  }

  // Free the memory pool
  rdmaMemoryFree();

  if (cqSize > 0)
  {
    if (ibv_destroy_cq(ibCQ) != 0)
      Errorm("ibv_destroy_cq failed");
    cqSize = 0;
  }
  if (ibv_dealloc_pd(ibPD) != 0)
    Errorm("ibv_dealloc_pd failed");
  if (ibv_destroy_comp_channel(ibCC) != 0)
    Errorm("ibv_destroy_comp_channel failed");
  if (ibv_close_device(ibContext) != 0)
    Error("ibv_close_device falied");
  ibContext = NULL;
  rdmaInitialized = false;
}


// Prepare for making an RDMA connection by creating a queue pair.
// Return queue pair number, location identifier, and memory key.
void RdmaConn::rdPrep(TcpConn *connP, UInt32 *qpnumP, UInt32 *lidP,
                      UInt32 *rkeyP)
{
  struct ibv_qp_init_attr qpIAttr;
  struct ibv_qp_attr qpAttr;
  struct ibv_port_attr portAttr;
  unsigned int n;
  int maxWr;
  int accFlags = IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ |
    IBV_ACCESS_REMOTE_WRITE;

  // Calculate the size of the work request queues.  Since RDMA writes are
  // initiated by tester threads sending from or reading into client nodes,
  // the total number of outstanding messages will never exceed the number
  // of testers times number of clients.  In addition to these messages,
  // worker threads could also be sending replies.  Double that number to
  // allow for the case where the RDMA receiver thread is handling a send
  // event (i.e. isn't actively polling), and wakes a tester or worker, who
  // then proceeds to send again before a new poll can be started,
  // overflowing the send queue.  If we make the queue big enough to allow
  // for all of these requests, it should never overflow.  The alternative
  // would be to manage the queuing ourselves, to limit the maximum number
  // of sends that are outstanding at one time.
  //
  // Make the receive queue the same size as the send queue.  That way
  // every send will be guaranteed to have a corresponding receive queue
  // entry available.
  maxWr = (nTesterThreads * nClients + nWorkers) * 2;

  // Cut back the size if it exceeds the device limit.
  if (maxWr > ibAttr.max_qp_wr)
    maxWr = ibAttr.max_qp_wr;

  // Create a queue pair
  memset(&qpIAttr, 0, sizeof(qpIAttr));
  qpIAttr.send_cq = ibCQ;
  qpIAttr.recv_cq = ibCQ;
  qpIAttr.sq_sig_all = 1;
  qpIAttr.qp_type = IBV_QPT_RC;
  qpIAttr.cap.max_send_wr = maxWr;
  qpIAttr.cap.max_recv_wr = (useRdma == rAll || useRdma == rInline) ? maxWr : 1;
  qpIAttr.cap.max_send_sge = (useRdma == rInline) ? 2 : 1;
  qpIAttr.cap.max_recv_sge = (useRdma == rInline) ? 2 : 1;
  qp = ibv_create_qp(ibPD, &qpIAttr);
  if (qp == NULL) Error("ibv_create_qp failed");

  // Get port attributes
  if (ibv_query_port(ibContext, rdmaPort, &portAttr) != 0)
    Errorm("ibv_query_port failed");

  *qpnumP = qp->qp_num;
  *lidP = portAttr.lid;
  *rkeyP = ibMR->rkey;

  // Change queue pair from RESET to INIT state
  memset(&qpAttr, 0, sizeof(qpAttr));
  qpAttr.qp_state = IBV_QPS_INIT;
  qpAttr.qp_access_flags = accFlags;
  qpAttr.pkey_index = 0;
  qpAttr.port_num = rdmaPort;

  if (ibv_modify_qp(qp, &qpAttr, IBV_QP_STATE | IBV_QP_ACCESS_FLAGS |
                    IBV_QP_PKEY_INDEX | IBV_QP_PORT) != 0)
    Errorm("ibv_modify_qp to INIT failed");

  // Post some receive requests
  if (useRdma == rAll || useRdma == rInline)
    for (n = 0; n < nTesterThreads; n++)
    {
      PollWait *pwaitP = new PollWait(connP);
      pwList.push_back(pwaitP);
      rdPostRecv(pwaitP);
    }
}


// Post a receive work request, reading into the mbuf in the
// PollWait object.
void RdmaConn::rdPostRecv(PollWait *pwaitP)
{
  struct ibv_sge sge;
  struct ibv_recv_wr rr;
  struct ibv_recv_wr *bad_wr;

  memset(&sge, 0, sizeof(sge));
  sge.addr = reinterpret_cast<UInt64>(pwaitP->mbufP);
  sge.length = mbufSize;
  sge.lkey = ibMR->lkey;

  memset(&rr, 0, sizeof(rr));
  rr.wr_id = reinterpret_cast<UInt64>(pwaitP);
  rr.next = NULL;
  rr.sg_list = &sge;
  rr.num_sge = 1;

  if (ibv_post_recv(qp, &rr, &bad_wr) != 0)
    Errorm("ibv_post_recv failed");
  else if (debugLevel > 1)
    Logt("RDMA read started, pwait " << pwaitP);
}


// If this is a client, establish an RDMA connection to the target server.
// For servers, set up RDMA for listening.
string RdmaConn::rdConnect(string destName, UInt32 qpnum,
                           UInt32 lid, UInt32 key)
{
  string errmsg;
  struct ibv_qp_attr qpAttr;
  DataBuff db;

  rqpnum = qpnum;
  rlid = lid;
  rkey = key;

  // Change queue pair to RTR state
  memset(&qpAttr, 0, sizeof(qpAttr));
  qpAttr.qp_state = IBV_QPS_RTR;
  qpAttr.path_mtu = IBV_MTU_2048;
  qpAttr.dest_qp_num = rqpnum;
  qpAttr.rq_psn = 0;
  qpAttr.max_dest_rd_atomic = 1;
  qpAttr.min_rnr_timer = 0x12;
  qpAttr.ah_attr.is_global = 0;
  qpAttr.ah_attr.dlid = rlid;
  qpAttr.ah_attr.sl = 0;
  qpAttr.ah_attr.src_path_bits = 0;
  qpAttr.ah_attr.port_num = rdmaPort;

  if (ibv_modify_qp(qp, &qpAttr,
                    IBV_QP_STATE | IBV_QP_AV | IBV_QP_PATH_MTU |
                    IBV_QP_DEST_QPN | IBV_QP_RQ_PSN |
                    IBV_QP_MAX_DEST_RD_ATOMIC | IBV_QP_MIN_RNR_TIMER) != 0)
    Errorm("ibv_modify_qp to RTR failed");

  // Change queue pair to RTS state
  memset(&qpAttr, 0, sizeof(qpAttr));
  qpAttr.qp_state = IBV_QPS_RTS;
  qpAttr.timeout = 0x12;
  qpAttr.retry_cnt = 6;
  qpAttr.rnr_retry = 0;
  qpAttr.sq_psn = 0;
  qpAttr.max_rd_atomic = 1;

  if (ibv_modify_qp(qp, &qpAttr,
                    IBV_QP_STATE | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT |
                    IBV_QP_RNR_RETRY | IBV_QP_SQ_PSN |
                    IBV_QP_MAX_QP_RD_ATOMIC) != 0)
    Errorm("ibv_modify_qp to RTS failed");

  if (debugLevel > 0)
    Logt("RDMA connected to " << destName);
  return errmsg;
}


// Tear down RDMA connection
void RdmaConn::rdDisconnect()
{
  if (ibv_destroy_qp(qp) != 0) Errorm("ibv_destroy_qp failed");
  while (!pwList.empty())
  {
    delete(pwList.front());
    pwList.pop_front();
  }
}


// Write data from test buffer to specified remote address using RDMA
void RdmaConn::rdWrite(DataBuff *testBuffP, RdmaAddr raddr, UInt32 rlen,
                       PollWait *pwaitP)
{
  struct ibv_sge sge;
  struct ibv_send_wr sr;
  struct ibv_send_wr *bad_wr;
  enum ibv_wc_status status;

  if (rlen != buffsize)
    Error("incorect length in RDMA write");

  memset(&sge, 0, sizeof(sge));
  sge.addr = reinterpret_cast<UInt64>(testBuffP->getBuffP());
  sge.length = testBuffP->getBufflen();
  sge.lkey = ibMR->lkey;

  memset(&sr, 0, sizeof(sr));
  sr.wr_id = reinterpret_cast<UInt64>(pwaitP);
  sr.next = NULL;
  sr.sg_list = &sge;
  sr.num_sge = 1;
  sr.opcode = IBV_WR_RDMA_WRITE;
  sr.send_flags = IBV_SEND_SIGNALED;
  sr.wr.rdma.remote_addr = raddr;
  sr.wr.rdma.rkey = rkey;

  if (ibv_post_send(qp, &sr, &bad_wr) != 0)
    Errorm("ibv_post_send failed in rdWrite");
  else if (debugLevel > 1)
    Logt("RDMA write started, pwait " << pwaitP);

  // Wait for RDMA receiver thread to notify us that operation is complete
  status = pwaitP->wait();
  if (status != IBV_WC_SUCCESS)
    Error("RDMA write failed, status = " << status);
}


// Read data from remote address using RDMA into the specified local
// buffer, which must be in registered memory.
void RdmaConn::rdRead(RdmaAddr raddr, UInt32 rlen, char *dataP,
                      PollWait *pwaitP)
{
  struct ibv_sge sge;
  struct ibv_send_wr sr;
  struct ibv_send_wr *bad_wr;
  enum ibv_wc_status status;

  if (rlen != buffsize)
    Error("incorect length in RDMA read");

  memset(&sge, 0, sizeof(sge));
  sge.addr = reinterpret_cast<UInt64>(dataP);
  sge.length = rlen;
  sge.lkey = ibMR->lkey;

  memset(&sr, 0, sizeof(sr));
  sr.wr_id = reinterpret_cast<UInt64>(pwaitP);
  sr.next = NULL;
  sr.sg_list = &sge;
  sr.num_sge = 1;
  sr.opcode = IBV_WR_RDMA_READ;
  sr.send_flags = IBV_SEND_SIGNALED;
  sr.wr.rdma.remote_addr = raddr;
  sr.wr.rdma.rkey = rkey;

  if (ibv_post_send(qp, &sr, &bad_wr) != 0)
    Errorm("ibv_post_send failed in rdRead");
  else if (debugLevel > 1)
    Logt("RDMA read started, pwait " << pwaitP);

  // Wait for RDMA receiver thread to notify us that operation is complete
  status = pwaitP->wait();
  if (status != IBV_WC_SUCCESS)
    Error("RDMA read failed, status = " << status);
}


// Send a message using the RDMA interface.  The data must be in a
// registered memory buffer.  An extra data buffer can be sent using
// auxBuffP in DataBuff.
void RdmaConn::rdSend(DataBuff *dbP, PollWait *pwaitP)
{
  struct ibv_sge sge[2];
  struct ibv_send_wr sr;
  struct ibv_send_wr *bad_wr;
  enum ibv_wc_status status;

  memset(&sr, 0, sizeof(sr));
  sr.wr_id = reinterpret_cast<UInt64>(pwaitP);
  sr.next = NULL;
  sr.sg_list = sge;
  sr.num_sge = 1;
  sr.opcode = IBV_WR_SEND;
  sr.send_flags = IBV_SEND_SIGNALED;

  memset(&sge, 0, sizeof(sge));
  sge[0].addr = reinterpret_cast<UInt64>(dbP->getBuffP());
  sge[0].length = dbP->getBufflen();
  sge[0].lkey = ibMR->lkey;

  if (dbP->getAuxlen() != 0)
  {
    sge[1].addr = reinterpret_cast<UInt64>(dbP->getAuxBuffP());
    sge[1].length = dbP->getAuxlen();
    sge[1].lkey = ibMR->lkey;
    sr.num_sge = 2;
  }

  if (ibv_post_send(qp, &sr, &bad_wr) != 0)
    Errorm("ibv_post_send failed in rdSend");
  else if (debugLevel > 1)
    Logt("RDMA send started, pwait " << pwaitP);

  // Wait for RDMA receiver thread to notify us that operation is complete
  status = pwaitP->wait();
  if (status != IBV_WC_SUCCESS)
    Error("RDMA send failed, status = " << status);
}


// Handle an incoming receive completion from the RDMA queue
void RdmaConn::rdRecv(PollWait *pwaitP, unsigned int len)
{
  DataBuff db;
  RcvMsg *rmsgP = new RcvMsg(pwaitP->connP);

  if (len < MSG_HDRSIZE)
    Error("RDMA message too short");

  db.initBuff(pwaitP->mbufP, len);
  if (db.getUInt32() != MSG_MAGIC)
    Error("Invalid message from " << pwaitP->connP->destName());
  rmsgP->msgId   = static_cast<MsgId>(db.getUInt32());
  rmsgP->msgType = static_cast<MType>(db.getUInt32());
  if (len > MSG_HDRSIZE)
  {
    rmsgP->msgBuff.newBuff(len - MSG_HDRSIZE);
    memcpy(rmsgP->msgBuff.getBuffP(), pwaitP->mbufP + MSG_HDRSIZE,
           len - MSG_HDRSIZE);
  }

  // Re-post the receive before dispatching the RcvMsg so that the thread
  // on the other node won't send us another message before we have an
  // mbuf ready to receive it.
  rdPostRecv(pwaitP);
  pwaitP->connP->gotMsg(rmsgP);
}


// PollWait constructor for posting send and receive requests
PollWait::PollWait(TcpConn *tconnP)
{
  init();
  mbufP = mbufGet(mbufSize);
  connP = tconnP;
}


// Initialize PollWait object
void PollWait::init()
{
  thInitMutex(&pwMutex);
  thInitCond(&pwCond);
  complete = false;
  status = IBV_WC_SUCCESS;
  mbufP = NULL;
  connP = NULL;
}


// PollWait Destructor
PollWait::~PollWait()
{
  if (mbufP != NULL)
    mbufFree(mbufP);
}


// Wait for wakeup
ibv_wc_status PollWait::wait()
{
  thLock(&pwMutex);
  while (!complete)
    thWait(&pwCond, &pwMutex);
  complete = false;
  thUnlock(&pwMutex);
  return status;
}


// Wake up waiter
void PollWait::wakeup(ibv_wc_status s)
{
  thLock(&pwMutex);
  status = s;
  complete = true;
  thSignal(&pwCond);
  thUnlock(&pwMutex);
}


#else
static string rdmaStart()  { return "RDMA not supported"; }
static void rdmaShutdown() {}
void RdmaConn::rdPrep(TcpConn *connP, UInt32 *qpnumP, UInt32 *lidP,
                      UInt32 *rkeyP) { *qpnumP = *lidP = *rkeyP = 0; }
string RdmaConn::rdConnect(string destName, UInt32 qpnum,
                           UInt32 lid, UInt32 key) { return "?"; }
void RdmaConn::rdDisconnect() {}
void RdmaConn::rdWrite(DataBuff *testBuffP, RdmaAddr raddr, UInt32 rlen,
                       PollWait *pwaitP) {}
void RdmaConn::rdRead(RdmaAddr raddr, UInt32 rlen, char *dataP,
                      PollWait *pwaitP) {}
void RdmaConn::rdSend(DataBuff *dbP, PollWait *pwaitP) {}
void RdmaConn::rdRecv(PollWait *pwaitP, unsigned int len) {}

static char *poolGet(unsigned int len) { return new char[len]; }
static void poolFree(char *buffP) { delete [] buffP; }
static char *mbufGet(unsigned int len) { return NULL; }
static void mbufFree(char *mbufP) {}
#endif // RDMA


// Wrapper for thread body
extern "C" void *thread_wrapper(void *argP)
{
  int rc;
  Thread *thP = static_cast<Thread *>(argP);
  thP->startup();
  rc = thP->threadBody();

  // Tell the main thread to clean us up
  thLock(&globalMutex);
  deadThreads.push_back(thP);
  thBcast(&globalCond);
  thUnlock(&globalMutex);
  pthread_exit(&rc);
  return NULL;
}


// Start the thread and wait for it to get going (it would be nice to do
// this from the constructor, except that virtual functions always use the
// base class version, so we wouldn't be able to call the thread body
// routine).
void Thread::init()
{
  thLock(&globalMutex);
  running = false;
  nThreadsStarted++;
  if (pthread_create(&th, NULL, thread_wrapper, this) != 0)
    Errorm("pthread_create");
  while (!running)
    thWait(&globalCond, &globalMutex);
  thUnlock(&globalMutex);
}


// Call this routine when a new thread starts to tell the init routine
// that the thread is running.
void Thread::startup()
{
  thLock(&globalMutex);
  running = true;
  thBcast(&globalCond);
  thUnlock(&globalMutex);
}


// Wait for thread to go away
void Thread::waitForDeath()
{
  thLock(&globalMutex);
  while (running)
    thWait(&globalCond, &globalMutex);
  thUnlock(&globalMutex);
}


// Receiver constructor
Receiver::Receiver()
{
  thInitMutex(&receiverMutex);
  thInitCond(&receiverCond);

  // Create a socket pair to be used for waking up receiver thread when it
  // is in poll.
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, rcvSocks) < 0)
    Errorm("socketpair");
  setSockNonblocking(rcvSocks[0]);
  setSockNonblocking(rcvSocks[1]);
}


// Receiver destructor
Receiver::~Receiver()
{
  int j;
  for (j = 0; j < 2; j++)
    if (close(rcvSocks[j]) < 0)
      Errorm("close rcvSocks");
}


// Receiver thread body
int Receiver::threadBody()
{
  int j, numfd, nFDs = 0;
  TcpConn *connP;
  bool isSockpair;
#ifdef USE_EPOLL
  int epollFd;
  struct epoll_event ev, events[MAX_POLLFD_NUM];
#else
  Sock socket = INVALID_SOCK;
  PollSock fdList[MAX_POLLFD_NUM];
  map<Sock, TcpConn *> socketTab;
  map<Sock, TcpConn *>::iterator siter;
  bool fdNeedsRefresh = false;
#endif

#ifdef USE_EPOLL
  epollFd = epoll_create(MAX_POLLFD_NUM);
  if (epollFd < 0)
    Errorm("epoll_create failed");

  ev.events = EPOLLIN | EPOLLET;
  ev.data.ptr = NULL;
  if (epoll_ctl(epollFd, EPOLL_CTL_ADD, rcvSocks[0], &ev) < 0)
    Errorm("epoll_ctl EPOLL_CTL_ADD rcvSocks");
#endif

  while (true)
  {
    // Wait for the listen thread to give us a socket
    thLock(&receiverMutex);
    while (newSockets.empty())
      thWait(&receiverCond, &receiverMutex);
    thUnlock(&receiverMutex);

    // Loop, polling the sockets
    while (true)
    {
      // If the ListenAccept thread passed us any new sockets, add them
      // to table of sockets that we poll.
      thLock(&receiverMutex);
      while (!newSockets.empty())
      {
        NewSock ns = newSockets.front();
        newSockets.pop_front();
#ifdef USE_EPOLL
        ev.events = EPOLLIN | EPOLLET;
        ev.data.ptr = ns.connP;
        if (epoll_ctl(epollFd, EPOLL_CTL_ADD, ns.sock, &ev) < 0)
          Errorm("epoll_ctl EPOLL_CTL_ADD");
        nFDs++;
#else
        if (socketTab.find(ns.sock) != socketTab.end())
          Error("socket " << ns.sock << " is already in socketTab");
        socketTab[ns.sock] = ns.connP;
        fdNeedsRefresh = true;
#endif
      }
      thUnlock(&receiverMutex);

#ifdef USE_EPOLL
      if (nFDs == 0)
        break;
#else
      // Rebuild poll array if socket table has changed
      if (fdNeedsRefresh)
      {
        if (socketTab.empty())
          break;
        memset(fdList, 0, sizeof(fdList));
        fdList[0].fd = rcvSocks[0];
        fdList[0].events = POLLIN;
        fdList[0].revents = 0;
        nFDs = 1;
        for (siter = socketTab.begin();
             siter != socketTab.end();
             ++siter, nFDs++)
        {
          if (nFDs >= MAX_POLLFD_NUM)
            Error("more than " << MAX_POLLFD_NUM << " connections per receiver");
          fdList[nFDs].fd = siter->first;
          fdList[nFDs].events = POLLIN;
          fdList[nFDs].revents = 0;
        }
        fdNeedsRefresh = false;
      }
#endif

      // Wait for data from a socket
      while (true)
      {
#ifdef USE_EPOLL
        numfd = epoll_wait(epollFd, events, MAX_POLLFD_NUM, 5 * 1000);
#else
        numfd = poll(fdList, nFDs, 5 * 1000);
#endif
        if (numfd >= 0)
          break;
        if (errno != EINTR)
          Errorm("poll");
      }

      for (j = 0; j < nFDs && numfd > 0; j++)
      {
#ifdef USE_EPOLL
        connP = reinterpret_cast<TcpConn *>(events[j].data.ptr);
        isSockpair = (connP == NULL);
#else
        if (fdList[j].revents == 0)
          continue;
        socket = fdList[j].fd;
        isSockpair = (socket == rcvSocks[0]);
#endif
        numfd--;

        // If this is the socket pair, that means someone sent us a wakeup
        // notification.  Remove all data from the socket.
        if (isSockpair)
        {
          char tmpBuf[16];
          while (recv(rcvSocks[0], tmpBuf, sizeof(tmpBuf), 0) > 0);
          continue;
        }

#ifndef USE_EPOLL
        siter = socketTab.find(socket);
        if (siter == socketTab.end())
          Error("socket " << socket << " not found in socketTab");
        connP = siter->second;
#endif

        // Handle event from socket.  Shut it down if error occurs.
        if (connP->receiverEvent() != E_OK)
        {
#ifdef USE_EPOLL
          if (epoll_ctl(epollFd, EPOLL_CTL_DEL, connP->getSock(), NULL) < 0)
            Errorm("epoll_ctl EPOLL_CTL_DEL");
          nFDs--;
#else
          socketTab.erase(connP->getSock());
          fdNeedsRefresh = true;
#endif
          connP->connShutdown();
          connP->releaseConn();
        }
      }
    }
  }
#ifdef USE_EPOLL
  if (close(epollFd) < 0)
    Errorm("close epollFd");
#endif
  return 0;
}


// Add a new TCP connection for this receiver to handle
void Receiver::addConn(Sock sock, TcpConn *connP)
{
  connP->holdConn();
  thLock(&receiverMutex);
  newSockets.push_back(NewSock(sock, connP));
  thSignal(&receiverCond);
  thUnlock(&receiverMutex);
  wakeUp();
}


// Wake up receiver thread if it is in poll
void Receiver::wakeUp()
{
  int rc;
  char buf[1] = {'x'};
  while (true)
  {
    rc = send(rcvSocks[1], buf, 1, 0);
    if (rc == 0)
      Error("rcvSocks disconnected");
    if (rc > 0 || errno == EAGAIN || errno == EWOULDBLOCK)
      break;
    if (errno != EINTR)
      Errorm("send to rcvSocks");
  }
}


// Perform round-robin selection of receiver threads
static Receiver *pickReceiver()
{
  Receiver *rcvP;
  thLock(&globalMutex);
  rcvP = *nextReceiver;
  if (++nextReceiver == receiverTab.end())
    nextReceiver = receiverTab.begin();
  thUnlock(&globalMutex);
  return rcvP;
}


// InuseWaiter constructor
InuseWaiter::InuseWaiter(MType tmt, MsgId tmsgId, unsigned int tdatalen)
{
  thInitCond(&iwCond);
  mt = tmt;
  msgId = tmsgId;
  datalen = tdatalen;
}


// Assign priority according to message type.  Messages with high-valued
// types will be granted use of the socket for sending before lower ones.
// Inbound messages with priority less than 5 are considered bulk messages,
// and will be processed after all other message types.
static int msgPriority(MType mt)
{
  int j;
  static int mpri[mtLast];
  static bool inited = false;

  if (mt <= 0 || mt >= mtLast)
    Error("Invalid message type");
  if (!inited)
  {
    // Set up message priority table for faster lookups
    for (j = 0; j < mtLast; j++)
      switch (j)
      {
        case mtKill:    mpri[j] = 9; break;
        case mtStatOn:  mpri[j] = 8; break;
        case mtStatOff: mpri[j] = 8; break;
        case mtWrite:   mpri[j] = 1; break;
        case mtRead:    mpri[j] = 1; break;
        case mtNwrite:  mpri[j] = 1; break;
        case mtGetdata: mpri[j] = 1; break;
        default:        mpri[j] = 5; break;
      }
    inited = true;
  }
  return mpri[mt];
}


// Convert message type to string for debug output
static string mtToString(MType mt)
{
  switch (mt)
  {
    case mtUnknown:         return "Unknown";
    case mtReply:           return "Reply";
    case mtReplyErr:        return "ReplyErr";
    case mtVersion:         return "Version";
    case mtWrite:           return "Write";
    case mtRead:            return "Read";
    case mtNwrite:          return "Nwrite";
    case mtGetdata:         return "Getdata";
    case mtKill:            return "Kill";
    case mtConnect:         return "Connect";
    case mtReset:           return "Reset";
    case mtRdmaConn:        return "RdmaConn";
    case mtRdmaDisconn:     return "RdmaDisconn";
    case mtRdmaWrite:       return "RdmaWrite";
    case mtParms:           return "Parms";
    case mtTest:            return "Test";
    case mtStatus:          return "Status";
    case mtStatOn:          return "StatOn";
    case mtStatOff:         return "StatOff";
    case mtIdlePct:         return "IdlePct";
    default:                break;
  }
  return "??";
}

   
// Compare two InuseWaiter entries to determine which has higher priority
bool InuseCmp::operator()(const InuseWaiter *w1P, const InuseWaiter *w2P) const
{
  // High priority messages go first
  int p1 = msgPriority(w1P->mt);
  int p2 = msgPriority(w2P->mt);
  if (p1 < p2) return true;
  if (p1 > p2) return false;

  // Short messages go ahead of long ones, but if they are close in size,
  // consider them to be the same.
  p1 = w1P->datalen / 1000;
  p2 = w2P->datalen / 1000;
  if (p1 < p2) return true;
  if (p1 > p2) return false;

  // Older messages go ahead of later ones
  return w1P->msgId < w2P->msgId;
}


// TcpConn constructor
TcpConn::TcpConn(Sock tsock, IpAddr tdest)
{
  thInitMutex(&connMutex);
  thInitCond(&connCond);
  refCount = 0;
  tcSock = tsock;
  dest = tdest;
  nextMsgId = 1;
  inuse = false;
  broken = false;
  thLock(&globalMutex);
  cnum = nextCnum++;
  thUnlock(&globalMutex);
  rconnP = NULL;
  recvState = rcv_idle;
  recvP = NULL;
  recvlen = 0;
  recvmax = 0;
  rmsgP = NULL;
}


// Put a hold on this TcpConn so that it won't be deleted
void TcpConn::holdConn()
{
  thLock(&connMutex);
  refCount++;
  if (debugLevel > 1)
    Log("holdConn " << refCount << ", sock " << tcSock);
  thUnlock(&connMutex);
}


// Release hold on this TcpConn and delete it if count is zero.  This is
// the only place that the socket is closed.
void TcpConn::releaseConn()
{
  thLock(&connMutex);
  refCount--;
  if (debugLevel > 1)
    Log("releaseConn " << refCount << ", sock " << tcSock);
  if (refCount > 0)
    thUnlock(&connMutex);
  else
  {
    if (!broken || rconnP != NULL)
      Error("TcpConn released while still connected");
    thUnlock(&connMutex);
    if (close(tcSock) < 0)
      Errorm("socket close");
    if (debugLevel > 0)
      Log("Closed connection to " << destName());
    delete this;
  }
}


// Shut down the TCP connection.  Everyone who is using the connection
// should wake up due to the socket shutdown and release their hold.  When
// the hold count goes to zero, the socket will be closed and the TcpConn
// deleted.  The broken flag ensures that only one thread calls shutdown
// and prevents new threads from starting to use the connection.
void TcpConn::connShutdown()
{
  thLock(&connMutex);
  if (!broken)
  {
    if (shutdown(tcSock, SHUT_RDWR) < 0 && errno != ENOTCONN)
      Errorm("socket shutdown");
    broken = true;

    // Wake up anyone who was waiting for exclusive use of the socket.
    // Wait for them to notice the broken connection and pop themselves off
    // the waiters list.
    while (!waiters.empty())
    {
      thSignal(&waiters.top()->iwCond);
      thWait(&connCond, &connMutex);
    }

    // Tear down any RDMA connection
    rdmaDisconnect();
  }
  thUnlock(&connMutex);

  // Remove any queued messages for this connection.  If a worker thread
  // already picked up the RcvMsg, then he should fail quickly since the
  // broken flag is set.
  thLock(&workerMutex);
  list<RcvMsg *>::iterator riter;
  list<RcvMsg *> *qP = &msgQueue;
  riter = qP->begin();
  while (true)
  {
    if (riter == qP->end())
    {
      if (qP != &msgQueue)
        break;
      qP = &bulkQueue;
      riter = qP->begin();
      continue;
    }
    
    RcvMsg *rmsgP = *riter;
    if (rmsgP->connP != this)
      ++riter;
    else
    {
      riter = qP->erase(riter);
      delete rmsgP;
    }
  }
  thUnlock(&workerMutex);

  // Scan the pending reply table and wake up anybody who is waiting for a
  // reply on this connection.
  thLock(&connMutex);
  map<MsgId, MsgRecord *>::iterator piter;
  for (piter = pendReplyTab.begin(); piter != pendReplyTab.end(); )
  {
    MsgId msgId = piter->first;
    MsgRecord *mrP = piter->second;
    pendReplyTab.erase(piter++);

    // Manufacture a dummy RcvMsg with errText set and use that to wake
    // up sender.
    RcvMsg *rmsgP = new RcvMsg(this);
    rmsgP->msgId = msgId;
    rmsgP->errText = "connection broken";
    mrP->gotReply(cnum, rmsgP);
  }
  thUnlock(&connMutex);
}


// Handle an event from poll
Errno TcpConn::receiverEvent()
{
  Errno err = E_OK;
  DataBuff db;

  if (debugLevel > 1)
    Logt("event on sock " << tcSock);

  // Keep reading until the socket blocks or is shut down due to an error
  while (true)
  {
    if (broken)
    {
      err = E_BROKEN;
      break;
    }

    switch (recvState)
    {
      case rcv_idle:
        rmsgP = new RcvMsg(this);
        recvP = rmsgP->hdr;
        recvmax = MSG_HDRSIZE;
        recvState = rcv_header;
        break;

      case rcv_header:
        err = recvMessage();
        if (err != E_OK)
          break;
        
        // Message header format:
        //    UInt32 magic      Magic number, should be MSG_MAGIC
        //    UInt32 msgId      Message identifier
        //    UInt32 msgType    Message type
        //    UInt32 msgLen     Length of data which follows this header

        db.initBuff(rmsgP->hdr, MSG_HDRSIZE);
        if (db.getUInt32() != MSG_MAGIC)
        {
          Log("Invalid message from " << destName());
          err = E_BADMSG;
          break;
        }
        rmsgP->msgId   = static_cast<MsgId>(db.getUInt32());
        rmsgP->msgType = static_cast<MType>(db.getUInt32());
        recvmax        = db.getUInt32();

        if (recvmax > 16 * 1024 * 1024)
          Error("Invalid msgLen");

        if (recvmax > 0)
        {
          rmsgP->msgBuff.newBuff(recvmax);
          recvP = rmsgP->msgBuffP();
        }
        recvState = rcv_data;
        break;

      case rcv_data:
        err = recvMessage();
        if (err != E_OK)
          break;

        gotMsg(rmsgP);
        rmsgP = NULL;
        recvState = rcv_idle;
        break;
        
      default:
        Error("invalid receiver state");
    }
    if (err == E_WOULDBLOCK)
      return E_OK;
    if (err != E_OK)
      break;
  }
  if (err != E_OK)
  {
    // Discard any partially received message
    delete rmsgP;
    rmsgP = NULL;
  }
  return err;
}


// Call this just after a message has been receieved, from either the
// TCP receiver or the RDMA receiver thread.
void TcpConn::gotMsg(RcvMsg *rmsgP)
{
  // If this is a reply, wake up the sender.  Otherwise, pass the
  // received message to a worker thread.
  if (rmsgP->msgType == mtReply || rmsgP->msgType == mtReplyErr)
  {
    MsgRecord *mrP;
    map<MsgId, MsgRecord *>::iterator piter;
    thLock(&connMutex);
    piter = pendReplyTab.find(rmsgP->msgId);
    if (piter == pendReplyTab.end())
      Error("reply not found in pending reply table");
    mrP = piter->second;
    pendReplyTab.erase(rmsgP->msgId);
    thUnlock(&connMutex);

    // If this is an error reply, pull out the error text
    if (rmsgP->msgType == mtReplyErr)
      rmsgP->errText = rmsgP->msgBuff.getString();

    mrP->gotReply(cnum, rmsgP);
  }
  else
  {
    thLock(&workerMutex);
    if (msgPriority(rmsgP->msgType) < 5)
      bulkQueue.push_back(rmsgP);
    else
      msgQueue.push_back(rmsgP);
    thSignal(&workerCond);
    thUnlock(&workerMutex);
  }
}


// Find the test buffer address used in an Nwrite message
void TcpConn::getSourceAddr(MsgId msgId, char **srcAddrPP,
                            unsigned int *srcLenP)
{
  MsgRecord *mrP;
  map<MsgId, MsgRecord *>::iterator piter;
  thLock(&connMutex);
  piter = pendReplyTab.find(msgId);
  if (piter == pendReplyTab.end())
    Error("pending message not found for Getdata request");
  mrP = piter->second;
  *srcAddrPP = mrP->srcAddrP;
  *srcLenP = mrP->srcLen;
  thUnlock(&connMutex);
}


// This routine is called from a receiver thread to read pending data.
// Data is read into the buffer pointed to by recvP until recvmax bytes
// have been read.  The recvlen variable is used to keep track of how much
// has been read so far.  Returns E_WOULDBLOCK if more data is needed from
// socket.
Errno TcpConn::recvMessage()
{
  while (recvlen < recvmax)
  {
    int rc = recv(tcSock, recvP, recvmax - recvlen, 0);
    if (rc == 0)
    {
      Log("Connection to " << destName() << " broken");
      return E_CONNRESET;
    }
    if (rc < 0 && errno != EAGAIN && errno != EWOULDBLOCK)
    {
      Logm("Receive from " << destName() << " failed");
      return E_CONNRESET;
    }
    if (rc < 0)
      return E_WOULDBLOCK;

    if (debugLevel > 1)
      Logt("received " << rc << " bytes");
    recvlen += rc;
    recvP += rc;
  }
  recvlen = 0;
  recvmax = 0;
  recvP = NULL;
  return E_OK;
}


// Send a message on this connection.  When a reply is received, it will be
// added to the reply list in mrP.  Caller must call waitForReplies before
// deleting the MsgRecord.
Errno TcpConn::sendMessage(MType mt, DataBuff *dbP, MsgRecord *mrP,
                           PollWait *pwaitP)
{
  return sendit(mt, mt, 0, dbP, mrP, pwaitP);
}


// Internal routine used by sendMessage and sendReply to send a message
Errno TcpConn::sendit(MType mt, MType origmt, MsgId msgId, DataBuff *mdbP,
                      MsgRecord *mrP, PollWait *pwaitP)
{
  union { char hdr[MSG_HDRSIZE]; UInt64 n; } h; // Make it double-word aligned
  DataBuff db;
  struct msghdr msg;
  struct iovec iov[2];
  int rc, iovCount;
  unsigned int skipBytes;
  PollSock ps;
  bool doRdma;
  char *dataP = NULL;
  unsigned int datalen = 0;

  if (mdbP != NULL)
  {
    dataP = mdbP->getBuffP();
    datalen = mdbP->getBufflen();
  }

  // If rdma setting is "all" or "inline", then send this message using the
  // RDMA interface rather than the TCP socket.  Only do this for message
  // types that are used in testing, not for administrative or setup
  // messages.
  doRdma = (useRdma == rAll || useRdma == rInline) &&
    (origmt == mtWrite || origmt == mtRead || origmt == mtNwrite ||
     origmt == mtGetdata || origmt == mtRdmaWrite);

  thLock(&connMutex);

  // Assign a unique message ID for this message
  if (msgId == 0)
    msgId = nextMsgId++;

  // Wait for exclusive use of the socket for sending
  if (!doRdma)
  {
    if (inuse)
    {
      InuseWaiter iw(origmt, msgId, datalen);
      waiters.push(&iw);
      while ((inuse && !broken) || waiters.top() != &iw)
        thWait(&iw.iwCond, &connMutex);
      waiters.pop();
      if (broken)
        thSignal(&connCond);
    }
    if (broken)
    {
      thUnlock(&connMutex);
      return E_BROKEN;
    }
    inuse = true;
  }

  if (mt != mtReply && mt != mtReplyErr)
  {
    // Record this message ID in the MsgRecord for caller to wait on.  We
    // must do this before adding an entry in pendReplyTab because a socket
    // break will scan pendReplyTab and attempt to remove the entry for
    // this ID from the MsgRecord.
    mrP->addMsg(cnum, msgId);

    // Add an entry to pending reply table, which will be used to wake the
    // MsgRecord when the reply comes in.  It will also be used to locate
    // waiters if the connection breaks.
    if (pendReplyTab.find(msgId) != pendReplyTab.end())
      Error("message " << msgId << " is already in pendReplyTab");
    pendReplyTab[msgId] = mrP;
  }
  thUnlock(&connMutex);

  if (debugLevel > 0)
    Logt("sending msg type " << mtToString(mt) << " ID " << msgId
         << " len " << datalen << " to " << destName());

  // Build message header.  If doing RDMA, put it in a registered message
  // buffer.  For inline RDMA, an extra buffer full of data may be sent in
  // the auxBuff, so pass that directly to rdmaSend in an auxBuff instead
  // of copying it.
  if (doRdma)
  {
    if (MSG_HDRSIZE + datalen > mbufSize)
      Error("RDMA message too big");

    // For RDMA messages, caller must supply a PollWait object with
    // a message buffer.
    if (pwaitP == NULL || pwaitP->mbufP == NULL)
      Error("PollWait missing in sendit");

    db.initBuff(pwaitP->mbufP, MSG_HDRSIZE + datalen);
    if (useRdma == rInline && mdbP != NULL)
      db.setAux(mdbP->getAuxBuffP(), mdbP->getAuxlen());
  }
  else
    db.initBuff(h.hdr, MSG_HDRSIZE);

  db.putUInt32(MSG_MAGIC);
  db.putUInt32(msgId);
  db.putUInt32(mt);
  db.putUInt32(datalen + db.getAuxlen());

  if (doRdma)
  {
    if (datalen > 0)
      memcpy(pwaitP->mbufP + MSG_HDRSIZE, dataP, datalen);
    rdmaSend(&db, pwaitP);
    return E_OK;
  }

  // Combine header and data into an iovec
  iov[0].iov_base = h.hdr;
  iov[0].iov_len = MSG_HDRSIZE;
  if (datalen == 0)
    iovCount = 1;
  else
  {
    iov[1].iov_base = dataP;
    iov[1].iov_len = datalen;
    iovCount = 2;
  }

  memset(&msg, 0, sizeof(msg));
  msg.msg_iov = iov;
  msg.msg_iovlen = iovCount;
  skipBytes = 0;

  // Send the header and data
  while (msg.msg_iovlen > 0)
  {
    if (broken)
      break;

    if (skipBytes > 0)
    {
      while (msg.msg_iovlen > 0 && skipBytes >= msg.msg_iov[0].iov_len)
      {
        skipBytes -= msg.msg_iov[0].iov_len;
        msg.msg_iov++;
        msg.msg_iovlen--;
      }
      if (skipBytes == 0)
        continue;
      msg.msg_iov[0].iov_base =
        static_cast<char *>(msg.msg_iov[0].iov_base) + skipBytes;
      msg.msg_iov[0].iov_len -= skipBytes;
      skipBytes = 0;
      continue;
    }

    rc = sendmsg(tcSock, &msg, SENDMSG_FLAGS);
    if (rc == 0)
    {
      Warn("connection to " << destName() << " broken");
      connShutdown();
      break;
    }
    if (rc < 0 && errno != EAGAIN && errno != EWOULDBLOCK)
    {
      Warnm("send to " << destName() << " failed");
      connShutdown();
      break;
    }
    if (rc > 0)
    {
      skipBytes = rc;
      if (debugLevel > 1)
        Logt("write " << rc << " bytes on sock " << tcSock);
      continue;
    }

    // Socket is blocked on send, so poll and try again
    ps.fd = tcSock;
    ps.events = POLLOUT;
    ps.revents = 0;
    while (true)
    {
      rc = poll(&ps, 1, -1);
      if (rc >= 0)
        break;
      if (errno != EINTR)
        Errorm("poll for send");
    }
    skipBytes = 0;
  }

  // Release exclusive use of the socket
  thLock(&connMutex);
  inuse = false;
  if (!waiters.empty())
    thSignal(&waiters.top()->iwCond);
  thUnlock(&connMutex);
  return broken ? E_BROKEN : E_OK;
}


// Establish an RDMA connection to the node on the other end of this
// TCP connection.  For clients, the remote connection info will be
// obtained by sending an mtRdmaConn to the server (rmsgP parameter
// will be NULL).  On the server, rmsgP will have the message from the
// client containing the connection info, and this routine will send a
// reply with our info.
string TcpConn::rdmaConnect(RcvMsg *rmsgP)
{
  string errmsg;
  RdmaConn *newRconnP;
  UInt32 qpnum, lid, key;
  DataBuff db;
  unsigned int n;
  MsgRecord mr;
  Errno err;
  char *buffP;
  bool clientSide = (rmsgP == NULL);

  if (rconnP != NULL)
    Error("RDMA already connected");

  // Prepare connection.  Fetch our connection info and put it into a
  // message buffer.
  newRconnP = new RdmaConn();
  newRconnP->rdPrep(this, &qpnum, &lid, &key);

  n = 3 * sizeof(UInt32);
  if (!clientSide)
    n += sizeof(UInt32) + nTesterThreads * sizeof(RdmaAddr);
  db.newBuff(n);
  db.putUInt32(qpnum);
  db.putUInt32(lid);
  db.putUInt32(key);

  if (clientSide)
  {
    // Tell the server to establish a connection back to us, and reply
    // with the connection info that we'll need.
    err = sendMessage(mtRdmaConn, &db, &mr, NULL);
    mr.waitForReplies();
    if (err != E_OK)
    {
      rmsgP = new RcvMsg(this);
      rmsgP->errText = "RDMA connection to " + destName() + " failed";
    }
    else
      rmsgP = mr.nextReply();

    if (!rmsgP->errText.empty())
    {
      errmsg = rmsgP->errText;
      delete rmsgP;
      newRconnP->rdDisconnect();
      delete newRconnP;
      return errmsg;
    }
  }

  // Get the remote connection info, either from the reply to our
  // mtRdmaConn message (client side) or from the data received in an
  // mtRdmaConn message (server side).
  qpnum = rmsgP->msgBuff.getUInt32();
  lid = rmsgP->msgBuff.getUInt32();
  key = rmsgP->msgBuff.getUInt32();

  errmsg = newRconnP->rdConnect(destName(), qpnum, lid, key);
  if (!errmsg.empty())
  {
    newRconnP->rdDisconnect();
    delete newRconnP;
    return errmsg;
  }

  if (clientSide)
  {
    // Get remote RDMA memory buffers for use in write tests
    for (n = rmsgP->msgBuff.getUInt32(); n > 0; n--)
      remoteBuffs.push_back(rmsgP->msgBuff.getUInt64());
    delete rmsgP;
  }
  else
  {
    // Give the client node some RDMA memory buffers to use in write tests.
    // Keep track of what we have given out so that we can return the
    // memory if the connection breaks.
    db.putUInt32(nTesterThreads);
    for (n = 0; n < nTesterThreads; n++)
    {
      buffP = poolGet(buffsize);
      db.putRdmaAddr(buffP);
      givenBuffs.push_back(buffP);
    }
    rmsgP->sendReply(&db);
  }
  rconnP = newRconnP;
  return "";
}


// Tear down RDMA connection
void TcpConn::rdmaDisconnect()
{
  if (rconnP == NULL)
    return;
  rconnP->rdDisconnect();
  rconnP = NULL;
  remoteBuffs.clear();
  while (!givenBuffs.empty())
  {
    poolFree(givenBuffs.front());
    givenBuffs.pop_front();
  }
}


// Write data from test buffer to specified remote address using RDMA
void TcpConn::rdmaWrite(DataBuff *testBuffP, RdmaAddr raddr, UInt32 rlen,
                        PollWait *pwaitP)
{
  if (raddr == 0)
    Error("RDMA write with no buffer address");
  if (rconnP == NULL)
    Error("RDMA write with no RDMA connection");
  rconnP->rdWrite(testBuffP, raddr, rlen, pwaitP);
}


// Read data from remote address using RDMA into the specified local buffer
void TcpConn::rdmaRead(RdmaAddr raddr, UInt32 rlen, char *dataP,
                       PollWait *pwaitP)
{
  if (raddr == 0)
    Error("RDMA read with no buffer address");
  if (rconnP == NULL)
    Error("RDMA read with no RDMA connection");
  rconnP->rdRead(raddr, rlen, dataP, pwaitP);
}


// Send a message using the RDMA interface
void TcpConn::rdmaSend(DataBuff *dbP, PollWait *pwaitP)
{
  if (rconnP == NULL)
    Error("RDMA write with no RDMA connection");
  rconnP->rdSend(dbP, pwaitP);
}


// Handle an incoming receive completion from the RDMA queue
void TcpConn::rdmaRecv(PollWait *pwaitP, unsigned int len)
{
  if (rconnP == NULL)
    Error("RDMA write with no RDMA connection");
  rconnP->rdRecv(pwaitP, len);
}


// Allocate an RDMA buffer from the pool on remote node
RdmaAddr TcpConn::getRemoteBuff()
{
  RdmaAddr rBuff;
  thLock(&connMutex);
  if (remoteBuffs.empty())
    Error("remote buffer pool exhausted");
  rBuff = remoteBuffs.front();
  remoteBuffs.pop_front();
  thUnlock(&connMutex);
  return rBuff;
}


// Return a remote RDMA buffer to the pool
void TcpConn::freeRemoteBuff(RdmaAddr rBuff)
{
  thLock(&connMutex);
  remoteBuffs.push_back(rBuff);
  thUnlock(&connMutex);
}


// Target destructor
Target::~Target()
{
  if (connP != NULL)
  {
    connP->connShutdown();
    connP->releaseConn();
  }
}


// Create a TCP connection to target node.  Return error message if failure.
string Target::makeConnection()
{
  string errmsg, ver;
  Sock sock;
  int rc;
  RcvMsg *rmsgP;

  if (connP != NULL)
    return errmsg;

  if ((sock = socket(iaddr.getFamily(), SOCK_STREAM, 0)) == INVALID_SOCK)
    Errorm("socket open");
  setSockSizes(sock);

  // Don't delay sending messages
  if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on)) < 0)
    Errorm("setsockopt TCP_NODELAY");

  // Make the connection
#ifdef IPV6_SUPPORT
  if (iaddr.fam == AT_IPV6)
  {
    struct sockaddr_in6 saddr;
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin6_family = iaddr.getFamily();
    saddr.sin6_port = htons(port);
    memcpy(&saddr.sin6_addr, iaddr.a, sizeof(saddr.sin6_addr));
    rc = connect(sock, reinterpret_cast<sockaddr *>(&saddr), sizeof(saddr));
  }
  else
#endif
  {
    struct sockaddr_in saddr;
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_family = iaddr.getFamily();
    saddr.sin_port = htons(port);
    memcpy(&saddr.sin_addr, iaddr.a, sizeof(saddr.sin_addr));
    rc = connect(sock, reinterpret_cast<sockaddr *>(&saddr), sizeof(saddr));
  }
  if (rc < 0)
  {
    errmsg = geterr(errno);
    close(sock);
    return errmsg;
  }

  // Set for non-blocking IO
  setSockNonblocking(sock);

  connP = new TcpConn(sock, iaddr);
  connP->holdConn();
  pickReceiver()->addConn(sock, connP);
  rmsgP = sendm(mtVersion);
  if (!rmsgP->errText.empty())
  {
    errmsg = rmsgP->errText;
    delete rmsgP;
    return errmsg;
  }

  // Verify that his version matches ours
  ver = rmsgP->msgBuff.getString();
  delete rmsgP;
  if (ver != version)
    return "Version mismatch.  Remote version is \"" + ver + "\".";

  Log("Connected to " << hostname);
  return errmsg;
}


// Send one message, wait for reply, and return it.  If an error occurs,
// put error message in errText.
RcvMsg *Target::sendm(MType mt, DataBuff *dbP, PollWait *pwaitP,
                      char *srcAddrP, unsigned int srcLen)
{
  Errno err;
  MsgRecord mr(srcAddrP, srcLen);
  err = connP->sendMessage(mt, dbP, &mr, pwaitP);
  mr.waitForReplies();
  if (err != E_OK)
  {
    RcvMsg *rmsgP = new RcvMsg(connP);
    rmsgP->errText = "connection to " + name() + " broken";
    return rmsgP;
  }
  return mr.nextReply();
}


// Convert host name and IP address to printable name for error messages
string Target::name() const
{
  return hostString(hostname, iaddr);
}


// MsgRecord constructor
MsgRecord::MsgRecord(char *tsrcAddrP, unsigned int tsrcLen)
{
  thInitMutex(&waitMutex);
  thInitCond(&waitCond);
  srcAddrP = tsrcAddrP;
  srcLen = tsrcLen;
}


// MsgRecord destructor
MsgRecord::~MsgRecord()
{
  thLock(&waitMutex);
  if (!waitTab.empty())
    Error("MsgRecord destructor called with messages pending");

  // Delete any replies
  while (!replies.empty())
  {
    delete replies.front();
    replies.pop_front();
  }
  thUnlock(&waitMutex);
}


// Add an entry to set of waiting messages
void MsgRecord::addMsg(int cnum, MsgId msgId)
{
  thLock(&waitMutex);
  if (!waitTab.insert(MsgKey(cnum, msgId)).second)
    Error("message " << msgId << " is already in waitTab");
  thUnlock(&waitMutex);
}


// Wait for replies from all messages to be received
void MsgRecord::waitForReplies()
{
  thLock(&waitMutex);
  while (!waitTab.empty())
    thWait(&waitCond, &waitMutex);
  thUnlock(&waitMutex);
}


// Wait for any outstanding replies, check all received replies for errors,
// printing a message if found, and then delete the replies.  Return true
// if an error was seen.
bool MsgRecord::checkReplies()
{
  RcvMsg *rmsgP;
  bool gotError = false;
  waitForReplies();
  for (rmsgP = nextReply(); rmsgP != NULL; rmsgP = nextReply())
  {
    if (rmsgP->showError()) gotError = true;
    delete rmsgP;
  }
  return gotError;
}


// Call this when a reply is received
void MsgRecord::gotReply(int cnum, RcvMsg *rmsgP)
{
  thLock(&waitMutex);
  if (waitTab.erase(MsgKey(cnum, rmsgP->msgId)) != 1)
    Error("msgId not found in waitTab");
  replies.push_back(rmsgP);
  if (waitTab.empty())
    thSignal(&waitCond);
  thUnlock(&waitMutex);
}


// Return the next reply, and remove it from reply list
RcvMsg *MsgRecord::nextReply()
{
  RcvMsg *rmsgP = NULL;
  thLock(&waitMutex);
  if (!replies.empty())
  {
    rmsgP = replies.front();
    replies.pop_front();
  }
  thUnlock(&waitMutex);
  return rmsgP;
}


// Add a new entry to histogram.  Input should be a response time in HTime
// units (nanoseconds).
void Histogram::addEntry(HTime t)
{
  if (t < 0)
    return;

  nEvents++;
  totalTime += t;

#ifdef LOG_HIST
  // Round to two significant digits to cut down the number of buckets.
  // This doesn't work very well, and requires changes to the plotting
  // code, so it is under ifdef for now.
  t = llrint(siground(t, 2));
#else
  // Round time to nearest millisecond to reduce the histogram size
  t = httomsec(t) * 1000000LL;
#endif

  map<HTime, UInt32>::iterator b = buckets.find(t);
  if (b == buckets.end())
    buckets[t] = 1;
  else
    b->second++;
}


// Add in the contents of another histogram
void Histogram::addHist(const Histogram *hP)
{
  map<HTime, UInt32>::const_iterator hiter;
  map<HTime, UInt32>::iterator b;
  for (hiter = hP->buckets.begin(); hiter != hP->buckets.end(); ++hiter)
  {
    b = buckets.find(hiter->first);
    if (b == buckets.end())
      buckets[hiter->first] = hiter->second;
    else
      b->second += hiter->second;
  }
  nEvents += hP->nEvents;
  totalTime += hP->totalTime;
}


// Print a histogram
ostream &operator<<(ostream &os, const Histogram &h)
{
  h.printHist(os);
  return os;
}


// Print a histogram
void Histogram::printHist(ostream &os) const
{
  int oldwidth;
  map<HTime, UInt32>::const_iterator b;

  if (nEvents == 0)
    return;

  oldwidth = os.width();
  for (b = buckets.begin(); b != buckets.end(); ++b)
    os << "  " << setw(7) << siground(httosec(b->first)*1000.0, 4)
       << "  " << setw(7) << b->second << endl;
  os.width(oldwidth);
}


// Return average value in histogram in seconds
double Histogram::average() const
{ return httosec(totalTime) / nEvents; }


// Return median value in histogram in seconds
double Histogram::median() const
{
  int n = 0;
  HTime retval = 0;
  map<HTime, UInt32>::const_iterator b;
  for (b = buckets.begin(); b != buckets.end(); ++b)
  {
    n += b->second;
    if (n >= nEvents/2)
    {
      retval = b->first;
      break;
    }
  }
  return httosec(retval);
}


// Return minimum time value in histogram in seconds
double Histogram::minTime() const
{ return buckets.empty() ? 0 : httosec(buckets.begin()->first); }


// Return maximum time value in histogram in seconds
double Histogram::maxTime() const
{ return buckets.empty() ? 0 : httosec(buckets.rbegin()->first); }


// Return maximum bucket value in this histogram
UInt32 Histogram::maxBucket() const
{
  if (buckets.empty())
    return 0;
  return max_element(buckets.begin(), buckets.end(), BucketCmp())->second;
}


// Send a reply to a message with data from the specified DataBuff.  If the
// DataBuff pointer is NULL, then send an empty reply.  If errText is not
// empty then send an error reply (mtReplyErr) with that text.
void RcvMsg::sendReply(DataBuff *dbP, string errText, PollWait *pwaitP)
{
  if (!errText.empty())
  {
    msgBuff.newBuff(calcLen(errText));
    msgBuff.putString(errText);
    connP->sendit(mtReplyErr, msgType, msgId, &msgBuff, NULL, pwaitP);
  }
  else
    connP->sendit(mtReply, msgType, msgId, dbP, NULL, pwaitP);
}


// If an error occurred on this reply, print a message and return true
bool RcvMsg::showError()
{
  if (errText.empty())
    return false;
  Log("Error reply from " << connP->destName() << ": " << errText);
  return true;
}


// Start handling a request from the admin node.  If another request is
// already being handled, send back an error reply and return true.  The
// admin node should never send a new request before the previous one has
// finished.  This routine is to catch the case of multiple admin nodes, or
// an admin node restart.
bool RcvMsg::startAdminReq()
{
  if (!server)
    Error("admin request received from " << connP->destName());
  thLock(&globalMutex);
  if (cmdInProgress)
  {
    thUnlock(&globalMutex);
    sendReply(NULL, "command already in progress");
    return true;
  }
  cmdInProgress = true;
  thUnlock(&globalMutex);
  return false;
}


// Stop handling an admin command request
void RcvMsg::endAdminReq()
{
  thLock(&globalMutex);
  cmdInProgress = false;
  thUnlock(&globalMutex);
}


// Message handler for mtVersion
//    Input:
//       None
//    Returns:
//       String version
void RcvMsg::handleVersion()
{
  msgBuff.newBuff(calcLen(version));
  msgBuff.putString(version);
  sendReply(&msgBuff);
}


// Message handler for mtWrite
//    Input:
//       char[buffsize]   Test data buffer
//    Returns:
//       UInt64    seed value found in test data (only used if verify is on)
void RcvMsg::handleWrite(MsgWorker *mwP)
{
  UInt64 seed = 0;
  if (msgBuff.getBufflen() >= sizeof(UInt64))
    seed = msgBuff.getUInt64();
  if (verify && !msgBuff.verifyBuff(seed))
    Error("Data verification failed");
  msgBuff.newBuff(sizeof(UInt64));
  msgBuff.putUInt64(seed);
  sendReply(&msgBuff, errText, &mwP->pwait);
}


// Message handler for mtRdmaWrite
//    Input:
//       RdmaAddr  RDMA buffer address where data was written
//       UInt32    buffsize
//    Returns:
//       UInt64    seed value found in test data (only used if verify is on)
void RcvMsg::handleRdmaWrite(MsgWorker *mwP)
{
#ifdef RDMA
  RdmaAddr raddr = msgBuff.getRdmaAddr();
  UInt32 rlen = msgBuff.getUInt32();
  DataBuff db(raddr, rlen);
  UInt64 seed = 0;

  // Be sure that the address passed is within our memory pool
  if (raddr.addr < memoryPoolBase.addr ||
      raddr.addr + rlen >
      memoryPoolBase.addr + static_cast<Int64>(poolBuffsize) * poolCount)
    Error("buffer address out of range in RdmaWrite");

  if (db.getBufflen() >= sizeof(UInt64))
    seed = db.getUInt64();
  if (verify && !db.verifyBuff(seed))
    Error("Data verification failed");
  msgBuff.newBuff(sizeof(UInt64));
  msgBuff.putUInt64(seed);
  sendReply(&msgBuff, errText, &mwP->pwait);
#else
  sendReply(NULL, "RDMA not supported");
#endif
}


// Message handler for mtNwrite
//    Input:
//       RdmaAddr  RDMA buffer address (not used if RDMA is off)
//       UInt32    buffsize
//    Returns:
//       UInt64    seed value found in test data (only used if verify is on)
void RcvMsg::handleNwrite(MsgWorker *mwP)
{
  string errText;
  DataBuff db;
  UInt64 seed = 0;

  // This is an NSD-style write request.  If RDMA is enabled, read directly
  // from the other node's test data buffer into this worker thread's
  // dedicated memory buffer.  Otherwise send a Getdata RPC to fetch the
  // data.
  if (useRdma != rOff)
  {
    RdmaAddr raddr = msgBuff.getRdmaAddr();
    UInt32 rlen = msgBuff.getUInt32();
    if (mwP->rdBuffP == NULL || mwP->rdLen != buffsize)
      Error("bad RDMA buffer in handleNwrite");
    if (rlen != mwP->rdLen)
      Error("incorect buffer size in handleNwrite");
    connP->rdmaRead(raddr, rlen, mwP->rdBuffP, &mwP->pwait);
    db.initBuff(mwP->rdBuffP, mwP->rdLen);
    if (db.getBufflen() >= sizeof(UInt64))
      seed = db.getUInt64();
    if (verify && !db.verifyBuff(seed))
      Error("Data verification failed");
  }
  else
  {
    MsgRecord mr;
    msgBuff.newBuff(sizeof(UInt32));
    msgBuff.putUInt32(msgId);
    Errno err = connP->sendMessage(mtGetdata, &msgBuff, &mr, &mwP->pwait);
    mr.waitForReplies();
    if (err != E_OK)
      errText = "send failed";
    else
    {
      RcvMsg *replyP = mr.nextReply();
      errText = replyP->errText;
      if (errText.empty())
      {
        if (replyP->msgBuff.getBufflen() > sizeof(UInt64))
          seed = replyP->msgBuff.getUInt64();
        if (verify && !replyP->msgBuff.verifyBuff(seed))
          Error("Data verification failed");
      }
      delete replyP;
    }
  }
  msgBuff.newBuff(sizeof(UInt64));
  msgBuff.putUInt64(seed);
  sendReply(&msgBuff, errText, &mwP->pwait);
}


// Message handler for mtGetdata
//    Input:
//       UInt32    MsgId of the Nwrite message
//    Returns:
//       char[buffsize]   Test data buffer
void RcvMsg::handleGetdata(MsgWorker *mwP)
{
  // Use the MsgId to find the MsgRecord of the original Nwrite message.
  // That will have the buffer address of the test buffer that we should
  // send.
  MsgId msgId = msgBuff.getUInt32();
  char *srcAddrP;
  unsigned int srcLen;
  connP->getSourceAddr(msgId, &srcAddrP, &srcLen);
  if (srcAddrP == NULL || srcLen == 0)
    Error("source address missing in Getdata request");
  msgBuff.initBuff(srcAddrP, srcLen);
  sendReply(&msgBuff);
}


// Message handler for mtRead
//    Input:
//       UInt64    seed for test data (only used if verify is on)
//       RdmaAddr  RDMA buffer address (0 if not using RDMA)
//       UInt32    buffsize
//    Returns:
//       Test data buffer (if not sent using RDMA)
void RcvMsg::handleRead(MsgWorker *mwP)
{
  // If verifying contents, generate test data according to specified seed.
  // Otherwise, send the same contents every time.
  UInt64 seed = msgBuff.getUInt64();
  if (verify)
    mwP->rtestBuff.fillBuff(seed);

  if (useRdma != rOff && useRdma != rInline)
  {
    RdmaAddr raddr = msgBuff.getRdmaAddr();
    UInt32 rlen = msgBuff.getUInt32();
    connP->rdmaWrite(&mwP->rtestBuff, raddr, rlen, &mwP->pwait);
    sendReply(NULL, errText, &mwP->pwait);
  }
  else
    sendReply(&mwP->rtestBuff, errText, &mwP->pwait);
}


// Message handler for mtConnect
//    Input:
//       UInt32 nServers
//       Array[nServers]
//         String hostname
//         IpAddr IP address (native host format)
//    Returns:
//       Nothing
void RcvMsg::handleConnect()
{
  int j, nServers = 0;
  IpAddr iaddr;
  string hname, errmsg;
  multimap<IpAddr, Target *>::iterator node;
  Target *targP;
  ostringstream os;

  if (startAdminReq())
    return;

  // Target nodes shouldn't have a client table; that's only for admin
  // nodes.
  if (!clientNodes.empty() || !allNodes.empty())
    Error("nodes table not empty");

  // Remove any broken connections from server table.  Mark the other nodes
  // inactive initially.
  for (node = serverNodes.begin(); node != serverNodes.end(); )
  {
    targP = node->second;
    if (targP->connP == NULL || targP->connP->isBroken())
    {
      delete targP;
      serverNodes.erase(node++);
    }
    else
    {
      targP->active = false;
      ++node;
    }
  }

  nServers = msgBuff.getUInt32();
  for (j = 0; j < nServers; j++)
  {
    hname = msgBuff.getString();
    iaddr = msgBuff.getIpAddr();

    // If this node is already in the server table, then we don't need to
    // connect to it here.  Mark it active so that it won't be deleted.
    // If we have multiple connections to the same IP address, only mark
    // the first "nParallel" connections active.
    pair<multimap<IpAddr, Target *>::iterator,
         multimap<IpAddr, Target *>::iterator> nodes;
    int nConnections = 0;
    nodes = serverNodes.equal_range(iaddr);
    for (node = nodes.first; node != nodes.second; ++node)
    {
      if (nConnections < nParallel)
        node->second->active = true;
      nConnections++;
    }
    if (nConnections > 0)
      continue;

    // This node isn't in table, so connect to it and add to table
    targP = new Target(hname, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
    {
      os << "Cannot connect to " << hname << ": " << errmsg << endl;
      delete targP;
    }
    else
      serverNodes.insert(pair<IpAddr, Target *>(iaddr, targP));
  }

  // Shut down any connections to servers that were not mentioned in the
  // message argument or exceeded the parallel count.
  for (node = serverNodes.begin(); node != serverNodes.end(); )
  {
    targP = node->second;
    if (targP->active)
      ++node;
    else
    {
      delete targP;
      serverNodes.erase(node++);
    }
  }

  // If more than one parallel connection was requested, make additional
  // connections if necessary.  They will look like extra serverNodes.
  if (nParallel > 1)
  {
    multimap<IpAddr, Target *>::const_iterator node;
    map<IpAddr, Target *> servers;
    map<IpAddr, Target *>::const_iterator srv;

    // Make a table of unique server IP addresses
    for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
      servers[node->second->iaddr] = node->second;

    // For each unique address, count the number of existing connections
    // and add more if necessary.
    for (srv = servers.begin(); srv != servers.end(); ++srv)
      for (j = serverNodes.count(srv->first); j < nParallel; j++)
      {
        targP = new Target(srv->second->hostname, srv->first);
        errmsg = targP->makeConnection();
        if (!errmsg.empty())
        {
          os << "Cannot make parallel connection to " << srv->second->hostname
             << ": " << errmsg << endl;
          delete targP;
        }
        else
          serverNodes.insert(pair<IpAddr, Target *>(srv->first, targP));
      }
  }

  // Make RDMA connections if necessary
  if (useRdma != rOff)
  {
    if (!rdmaInitialized)
      os << "RDMA is not initialized" << endl;
    else
      for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
      {
        targP = node->second;
        errmsg = targP->connP->rdmaConnect(NULL);
        if (!errmsg.empty())
          os << "Cannot make RDMA connection to " << targP->hostname
             << ": " << errmsg << endl;
      }
  }

  endAdminReq();
  sendReply(NULL, os.str());
}


// Message handler for mtReset
//    Input:
//       None
//    Returns:
//       Nothing
void RcvMsg::handleReset()
{
  multimap<IpAddr, Target *>::iterator node;

  if (startAdminReq())
    return;
  if (!clientNodes.empty() || !allNodes.empty())
    Error("nodes table not empty");
  for (node = serverNodes.begin(); node != serverNodes.end(); )
  {
    delete node->second;
    serverNodes.erase(node++);
  }
  endAdminReq();
  sendReply(NULL);
}


// Message handler for mtRdmaConn - set up server-side RDMA connection
//    Input:
//       UInt32 client's queue pair number
//       UInt32 client's location identifier
//       UInt32 client's receive buffer key
//    Returns:
//       UInt32 my queue pair number
//       UInt32 my location identifier
//       UInt32 my receive buffer key
//       UInt32 nEntries
//       nEntries * RdmaAddr - RDMA memory buffers to use in write tests
void RcvMsg::handleRdmaConn()
{
#ifdef RDMA
  string errmsg;

  if (ibContext == NULL)
    Error("RDMA connect request received, but RDMA is not initialized");

  errmsg = connP->rdmaConnect(this);
  if (!errmsg.empty())
    sendReply(NULL, errmsg);
#else
  sendReply(NULL, "RDMA not supported");
#endif
}


// Message handler for mtRdmaDisconn - tear down RDMA connections
//    Input:
//       None
//    Returns:
//       Nothing
void RcvMsg::handleRdmaDisconn()
{
#ifdef RDMA
  multimap<IpAddr, Target *>::iterator node;
  Target *targP;

  // If we're a client, disconnect all of the servers
  for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
  {
    targP = node->second;

    // Disconnect our side
    targP->connP->rdmaDisconnect();

    // Tell server to disconnect his side
    RcvMsg *rmsgP = targP->sendm(mtRdmaDisconn);
    rmsgP->showError();
    delete rmsgP;
  }

  // On server, just disconnect from client who sent us the message
  connP->rdmaDisconnect();

  sendReply(NULL);
#else
  sendReply(NULL, "RDMA not supported");
#endif
}


// Break any existing TCP connection for this target
static void disconnect(pair<IpAddr, Target *> p)
{
  Target *targP = p.second;
  if (targP->connP == NULL || targP->connP->isBroken())
    return;
  targP->connP->connShutdown();
  targP->connP->releaseConn();
  targP->connP = NULL;
}


// Message handler for mtParms
//    Input:
//       UInt32 testTime
//       UInt32 buffsize
//       UInt32 socksize
//       UInt32 nTesterThreads
//       UInt32 nParallel
//       UInt32 useRdma
//       UInt32 verify
//       UInt32 nClients
//    Returns:
//       Nothing
void RcvMsg::handleParms()
{
  int newSocksize;
  string errText;
  vector<MsgWorker *>::iterator worker;

  if (startAdminReq())
    return;

  testTime = msgBuff.getUInt32();
  buffsize = msgBuff.getUInt32();
  newSocksize = msgBuff.getUInt32();
  nTesterThreads = msgBuff.getUInt32();
  nParallel = msgBuff.getUInt32();
  useRdma = static_cast<RdmaMode>(msgBuff.getUInt32());
  verify = msgBuff.getUInt32() != 0;
  nClients = msgBuff.getUInt32();

  if (testTime < 1) testTime = 1;
  if (buffsize < 1) buffsize = 1;
  if (buffsize > MAX_BUFFSIZE) buffsize = MAX_BUFFSIZE;
  if (newSocksize > MAX_SOCKSIZE) newSocksize = MAX_SOCKSIZE;
  if (nTesterThreads < 1) nTesterThreads = 1;
  if (nTesterThreads > MAX_TESTERS) nTesterThreads = MAX_TESTERS;
  if (nParallel < 1) nParallel = 1;
  if (nParallel > MAX_PARALLEL) nParallel = MAX_PARALLEL;

  // If socket buffer size changes, close any existing server connections
  // so that they will be re-opened with the proper settings.  This will
  // also ensure that the server has found out the new socksize value and
  // used that setting on his end.  Also, tell the listen/accept thread to
  // switch to the new size.
  if (newSocksize != socksize)
  {
    socksize = newSocksize;
    for_each(serverNodes.begin(), serverNodes.end(), disconnect);
    if (laThreadP != NULL)
      laThreadP->updateSocksize();
  }
  
  // If any worker threads have memory buffers allocated, free them
  // in case the RDMA memory pool needs to be resized.
  for (worker = workerTab.begin(); worker != workerTab.end(); ++worker)
  {
    MsgWorker *w = *worker;
    if (w->rdBuffP != NULL)
      w->freeRdmaBuff();
    if (w->rtestBuff.getBuffP() != NULL)
    {
      poolFree(w->rtestBuff.getBuffP());
      w->rtestBuff.initBuff(NULL, 0);
    }
  }

  // Start up RDMA device, or shut it down
  if (useRdma != rOff)
  {
    // Initialize the RDMA device and memory pool
    errText = rdmaStart();

    // Give each message worker thread a dedicated RDMA buffer
    if (errText.empty())
      for (worker = workerTab.begin(); worker != workerTab.end(); ++worker)
        (*worker)->getRdmaBuff();
  }
  else
    rdmaShutdown();

  // Allocate a test data buffer for each worker to use for replying to
  // read requests.  If RDMA is enabled, this will be allocated from the
  // RDMA memory pool.
  for (worker = workerTab.begin(); worker != workerTab.end(); ++worker)
  {
    MsgWorker *w = *worker;
    w->rtestBuff.initBuff(poolGet(buffsize), buffsize);    
    w->rtestBuff.fillBuff(randSeed());
  }

  // If we don't have enough tester threads, start up some new ones.
  // If we have too many, wake them up so that some of them will exit.
  thLock(&testerMutex);
  while (testerTab.size() < nTesterThreads)
  {
    thUnlock(&testerMutex);
    Tester *tP = new Tester();
    tP->init();
    thLock(&testerMutex);
    if (!testerTab.insert(tP).second)
      Error("duplicate testerTab entry");
  }
  if (testerTab.size() > nTesterThreads)
  {
    thBcast(&testerCond);
    while (testerTab.size() > nTesterThreads)
      thWait(&testerCond, &testerMutex);
  }
  thUnlock(&testerMutex);

  endAdminReq();
  sendReply(NULL, errText);
}


// Message handler for mtTest
//    Input:
//       UInt32 testType
//    Returns:
//       UInt64 totBytes
//       UInt32 CPU idle percentage
//       Histogram response times
void RcvMsg::handleTest()
{
  TType ttin, tt;
  multimap<IpAddr, Target *>::const_iterator node;
  string errText;
  Histogram hist;

  if (startAdminReq())
    return;
  if (serverNodes.empty())
    errText = "no server nodes for mtTest";

  // Pick up test type from incoming message
  ttin = static_cast<TType>(msgBuff.getUInt32());
  if (ttin >= ttLast && errText.empty())
    errText = "invalid test type";

  if (errText.empty())
  {
    // Run the test.  Pass a TestReq to each tester thread, and wait for
    // them all to complete it.
    HTime endTime;
    UInt64 totBytes;
    TestReq *trP;
    set<Tester *>::iterator tst;

    thLock(&testerMutex);
    if (testerTab.empty())
      Error("no tester threads");
    endTime = getTime() + sectoht(testTime);

    // The serverNodes table is ordered by IP address, and might contain
    // more than one connection per address.  We want round-robin sends to
    // go to each server before repeating on parallel connections, so make
    // a new table of addresses with proper sort order.  Note that the
    // serverNodes table will not change in the middle of a test, so we
    // don't need a mutex to access it here.
    vector<SortedTarget> rrNodes;
    IpAddr lastAddr;
    int ndx = 0;

    rrNodes.reserve(serverNodes.size());
    for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
    {
      Target *targP = node->second;
      if (targP->iaddr != lastAddr)
      {
        lastAddr = targP->iaddr;
        ndx = 0;
      }
      rrNodes.push_back(SortedTarget(ndx, targP));
      ndx++;
    }
    sort(rrNodes.begin(), rrNodes.end(), SortedTarget::comp);
    vector<SortedTarget>::const_iterator rnode = rrNodes.begin();
    int nServers = rrNodes.size() / nParallel;

    for (tst = testerTab.begin(); tst != testerTab.end(); ++tst)
    {
      nTestersWorking++;

      // For the read/write test, half the workers are readers and half are
      // writers.
      if (ttin == ttRW)
        tt = nTestersWorking > testerTab.size() / 2 ? ttWrite : ttRead;
      else
        tt = ttin;
      trP = new TestReq(tt, endTime);

      // If test isn't round-robin, assign one server target.
      // Otherwise assign all of them.  With parallel sockets, give
      // separate sets of connections to each thread if possible.
      int ns = (tt == ttSwrite || tt == ttSread) ? 1 : nServers;
      for (ndx = 0; ndx < ns; ndx++)
      {
        // For RDMA write tests, reserve memory on other nodes to write
        // into.  We avoid the extra round trip required by NSD-style
        // writes by choosing the remote memory location on the sending
        // side, and then telling the target where to find the data.
        RdmaAddr tBuff;
        if (useRdma != rOff && (tt == ttWrite || tt == ttSwrite))
          tBuff = rnode->targP->connP->getRemoteBuff();

        trP->testNodes.push_back(TNodeInfo(rnode->targP, tBuff));
        rnode++;
        if (rnode == rrNodes.end())
          rnode = rrNodes.begin();
      }

      // Hand off the test request to a tester thread
      (*tst)->doTest(trP);

      // Return RDMA memory buffers if we acquired them
      if (useRdma != rOff && (tt == ttWrite || tt == ttSwrite))
      {
        list<TNodeInfo>::iterator tnode;
        for (tnode = trP->testNodes.begin();
             tnode != trP->testNodes.end();
             ++tnode)
          tnode->targP->connP->freeRemoteBuff(tnode->tBuff);
      }
    }
    thBcast(&testerCond);

    while (nTestersWorking > 0)
      thWait(&testerCond, &testerMutex);

    // Tally the results from the tester threads
    totBytes = 0;
    while (!doneList.empty())
    {
      trP = doneList.front();
      doneList.pop_front();
      totBytes += trP->totBytes;
      hist.addHist(&trP->hist);
      if (errText.empty())
        errText = trP->errText;
      delete trP;
    }
    thUnlock(&testerMutex);

    msgBuff.newBuff(sizeof(UInt64) + sizeof(UInt32) + hist.calcLen());
    msgBuff.putUInt64(totBytes);
    msgBuff.putUInt32(idleTime);
    hist.putBuff(&msgBuff);
  }

  endAdminReq();
  sendReply(&msgBuff, errText);
}


// Message handler for mtStatus
//    Input:
//       None
//    Returns:
//       String ServerStatus
void RcvMsg::handleStatus()
{
  ostringstream os;
  multimap<IpAddr, Target *>::iterator node;

  if (startAdminReq())
    return;
  for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
    os << "    -> " << node->second->name() << endl;

  string s = os.str();
  msgBuff.newBuff(calcLen(s));
  msgBuff.putString(s);
  endAdminReq();
  sendReply(&msgBuff);
}


// Message handler for mtIdlePct
//    Input:
//       None
//    Returns:
//       UInt32 CPU idle percentage
void RcvMsg::handleIdlePct()
{
  msgBuff.newBuff(sizeof(UInt32));
  msgBuff.putUInt32(idleTime);
  sendReply(&msgBuff);
}


// Dispatch a received message to the appropriate handler.  Simple messages
// are handled here.
void RcvMsg::dispatch(MsgWorker *mwP)
{
  if (debugLevel > 0)
    Logt("got msg type " << mtToString(msgType) << " ID " << msgId
         << " len " << msgLen() << " from " << connP->destName());

  switch (msgType)
  {
    case mtVersion:
      handleVersion();
      break;

    case mtKill:
      sendReply(NULL);
      connP->connShutdown();
      exit(EXIT_SUCCESS);
      break;

    case mtWrite:
      handleWrite(mwP);
      break;

    case mtRdmaWrite:
      handleRdmaWrite(mwP);
      break;

    case mtNwrite:
      handleNwrite(mwP);
      break;

    case mtGetdata:
      handleGetdata(mwP);
      break;

    case mtRead:
      handleRead(mwP);
      break;

    case mtConnect:
      handleConnect();
      break;

    case mtReset:
      handleReset();
      break;

    case mtRdmaConn:
      handleRdmaConn();
      break;

    case mtRdmaDisconn:
      handleRdmaDisconn();
      break;

    case mtParms:
      handleParms();
      break;

    case mtTest:
      handleTest();
      break;

    case mtStatus:
      handleStatus();
      break;

    case mtStatOn:
      collectStats = true;
      (void) cpuIdle();
      sendReply(NULL);
      break;

    case mtStatOff:
      collectStats = false;
      idleTime = cpuIdle();
      sendReply(NULL);
      break;

    case mtIdlePct:
      handleIdlePct();
      break;

    default:
      Error("Invalid message from " << connP->destName());
  }
}


// Message worker thread body
int MsgWorker::threadBody()
{
  RcvMsg *rmsgP;
  thLock(&workerMutex);
  while (true)
  {
    // Wait for a receiver thread to pass us a message.  Don't process bulk
    // messages if almost all of the workers are busy.
    while (msgQueue.empty() &&
           (bulkQueue.empty() || workersActive >= nWorkers - 2))
      thWait(&workerCond, &workerMutex);
    workersActive++;
    if (msgQueue.empty())
    {
      rmsgP = bulkQueue.front();
      bulkQueue.pop_front();
    }
    else
    {
      rmsgP = msgQueue.front();
      msgQueue.pop_front();
    }

    // Call the message handler
    thUnlock(&workerMutex);
    rmsgP->dispatch(this);
    delete rmsgP;
    thLock(&workerMutex);
    workersActive--;
  }
  thUnlock(&workerMutex);
  return 0;
}


// Reserve an RDMA memory buffer, message buffer, and test buffer for this
// worker thread.
void MsgWorker::getRdmaBuff()
{
  if (!rdmaInitialized || rdBuffP != NULL)
    Error("invalid call to getRdmaBuff");
  rdLen = buffsize;
  rdBuffP = poolGet(rdLen);
  pwait.mbufP = (useRdma == rAll || useRdma == rInline) ?
    mbufGet(mbufSize) : NULL;
}


// If an RDMA memory buffer was allocated, free it
void MsgWorker::freeRdmaBuff()
{
  if (rdBuffP == NULL)
    return;
  if (!rdmaInitialized)
    Error("invalid call to freeRdmaBuff");
  poolFree(rdBuffP);
  rdLen = 0;
  rdBuffP = NULL;
  mbufFree(pwait.mbufP);
  pwait.mbufP = NULL;
}


// Tester thread body
int Tester::threadBody()
{
  HTime now, lastTime;
  list<TNodeInfo>::iterator tnode;
  Target *targP;
  RdmaAddr tBuff;
  RcvMsg *replyP = NULL;
  unsigned int datalen = 0;
  char *dataP = NULL;
  DataBuff db, tdb, testBuff;
  PollWait pwait;
  UInt64 seed = randSeed();
  bool isRead;

  thLock(&testerMutex);
  while (true)
  {
    // Wait for a work request or a notification that too many threads
    // are running.
    while (reqP == NULL && testerTab.size() <= nTesterThreads)
      thWait(&testerCond, &testerMutex);
    if (testerTab.size() > nTesterThreads)
      break;
    thUnlock(&testerMutex);

    if (reqP->testNodes.empty())
      Error("No servers for tester thread");

    // Initialize test buffer for sends.  If RDMA is enabled, this buffer
    // will be allocated in registered memory so that RDMA can be done
    // directly from it.
    testBuff.initBuff(poolGet(buffsize), buffsize);
    testBuff.fillBuff(seed);

    // If using RDMA, get a registered memory buffer to use for reads
    isRead = (reqP->tt == ttRead || reqP->tt == ttSread);
    if (isRead)
    {
      if (useRdma != rOff)
      {
        datalen = buffsize;
        dataP = poolGet(datalen);
      }
      else
      {
        datalen = 0;
        dataP = NULL;
      }
    }

    // For RDMA NSD writes, we need to send the test buffer address so
    // that the other node can do an RDMA read from it.
    if (reqP->tt == ttNwrite)
    {
      tdb.newBuff(sizeof(RdmaAddr) + sizeof(UInt32));
      tdb.putRdmaAddr(testBuff.getBuffP());
      tdb.putUInt32(testBuff.getBufflen());
    }

    // Get a registered message buffer for sends, if necessary
    if (useRdma == rAll || useRdma == rInline)
      pwait.mbufP = mbufGet(mbufSize);

    // Move to a random starting position in the test nodes list
    thLock(&globalMutex);
    int spos = random();
    thUnlock(&globalMutex);
    tnode = reqP->testNodes.begin();
    for (spos = spos % reqP->testNodes.size(); spos > 0; spos--)
      if (++tnode == reqP->testNodes.end())
        tnode = reqP->testNodes.begin();

    // Run the test
    now = getTime();
    while (now < reqP->endTime)
    {
      lastTime = now;

      targP = tnode->targP;
      tBuff = tnode->tBuff;
      tnode++;
      if (tnode == reqP->testNodes.end())
        tnode = reqP->testNodes.begin();

      switch (reqP->tt)
      {
        case ttWrite:
        case ttSwrite:
          if (useRdma == rOff)
          {
            replyP = targP->sendm(mtWrite, &testBuff);
          }
          else if (useRdma == rInline)
          {
            // Send the data inline with the Write message.  Since the data
            // is already in a registered memory buffer, pass it using
            // auxBuff so that it won't be copied in sendit().
            db.initBuff(NULL, 0);
            db.setAux(testBuff.getBuffP(), testBuff.getBufflen());
            replyP = targP->sendm(mtWrite, &db, &pwait);
            db.setAux(NULL, 0);
          }
          else
          {
            // Write directly to remote buffer
            if (tBuff == 0)
              Error("remote RDMA buffer missing");
            targP->connP->rdmaWrite(&testBuff, tBuff, buffsize, &pwait);

            // Tell the other side where we wrote the data
            db.newBuff(sizeof(RdmaAddr) + sizeof(UInt32));
            db.putRdmaAddr(tBuff);
            db.putUInt32(buffsize);
            replyP = targP->sendm(mtRdmaWrite, &db, &pwait);
          }
          break;

        case ttNwrite:
          // Pass test buffer address into MsgRecord so that Getdata can
          // retrieve it.
          replyP = targP->sendm(mtNwrite, &tdb, &pwait, testBuff.getBuffP(),
                                testBuff.getBufflen());
          break;

        case ttRead:
        case ttSread:
          db.newBuff(sizeof(UInt64) + sizeof(RdmaAddr) + sizeof(UInt32));
          db.putUInt64(seed);
          db.putRdmaAddr(dataP);
          db.putUInt32(datalen);
          replyP = targP->sendm(mtRead, &db, &pwait);
          break;

        default:
          Error("invalid test type: " << reqP->tt);
          break;
      }
      now = getTime();
      if (!replyP->errText.empty())
      {
        reqP->errText = replyP->errText;
        delete replyP;
        if (dataP != NULL)
          poolFree(dataP);
        if (pwait.mbufP != NULL)
        {
          mbufFree(pwait.mbufP);
          pwait.mbufP = NULL;
        }
        break;
      }

      if (collectStats)
      {
        reqP->totBytes += buffsize + MSG_HDRSIZE;
        reqP->hist.addEntry(now - lastTime);
      }

      if (verify)
      {
        // For a reads, we tell the other node what seed to use to generate
        // the test data, and verify that is what we got.  For writes, the
        // other node picks up the seed from the first double-word of the
        // test buffer that we send, verifies the contents, and tells us
        // the seed.  It must much the seed that we used to generate the
        // buffer.
        if (isRead)
        {
          if (!replyP->msgBuff.verifyBuff(seed))
            Error("Data verification failed");
          seed += SCRAMBLE;
        }
        else
        {
          if (replyP->msgBuff.getUInt64() != seed)
            Error("Data verification error");

          // When verifying contents, generate new test data for each
          // message.  Don't use randSeed() here because we don't want to
          // contend for the global mutex.
          seed += SCRAMBLE;
          testBuff.fillBuff(seed);
        }
      }
      delete replyP;
    }
    reqP->endTime = now;

    // Report results to main thread
    thLock(&testerMutex);
    doneList.push_back(reqP);
    reqP = NULL;
    if (nTestersWorking == 0)
      Error("doneList overflow");
    nTestersWorking--;
    if (nTestersWorking == 0)
      thBcast(&testerCond);

    if (dataP != NULL)
    {
      poolFree(dataP);
      dataP = NULL;
    }
    if (pwait.mbufP != NULL)
    {
      mbufFree(pwait.mbufP);
      pwait.mbufP = NULL;
    }
    poolFree(testBuff.getBuffP());
  }

  if (testerTab.erase(this) != 1)
    Error("thread pointer not found in testerTab");
  thBcast(&testerCond);
  thUnlock(&testerMutex);
  return 0;
}


// Tell tester thread to do some work.  This doesn't wake up the thread.
// Caller must broadcast testerCond to do that.  Tester will put updated
// TestReq on doneList when finished.  Caller must hold testerMutex.
void Tester::doTest(TestReq *trP)
{
  if (reqP != NULL)
    Error("tester thread busy");
  reqP = trP;
}


// Listen/accept thread body
int ListenAccept::threadBody()
{
  int rc;
  Sock lsock, sock;
  struct linger ling;
#ifdef IPV6_SUPPORT
  struct sockaddr_in6 saddr6;
#endif
  struct sockaddr_in saddr4;
  struct sockaddr *saddrP;
  socklen_t saddrLen, alen;
  PollSock ps;
  IpAddr iaddr;

  // Loop, opening a listen socket
  while (true)
  {
    if ((lsock = socket(addrFamily, SOCK_STREAM, 0)) == INVALID_SOCK)
      Errorm("socket open"); 

    // Set to recycle addresses
    if ((setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))) < 0)
      Errorm("setsockopt SO_REUSEADDR");

    // Set the socket to not linger after the program exits
    ling.l_onoff = 1;
    ling.l_linger = 0;
    if (setsockopt(lsock, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling)) < 0)
      Errorm("setsockopt SO_LINGER");

    // Set send/receive buffer sizes
    setSockSizes(lsock);

    // Bind the socket
#ifdef IPV6_SUPPORT
    if (useipv6)
    {
      memset(&saddr6, 0, sizeof(saddr6));
      saddr6.sin6_family = addrFamily;
      saddr6.sin6_port = htons(port);
      saddr6.sin6_addr = in6addr_any;
      saddrP = reinterpret_cast<sockaddr *>(&saddr6);
      saddrLen = sizeof(saddr6);
    }
    else
#endif
    {
      memset(&saddr4, 0, sizeof(saddr4));
      saddr4.sin_family = addrFamily;
      saddr4.sin_port = htons(port);
      saddr4.sin_addr.s_addr = htonl(INADDR_ANY);
      saddrP = reinterpret_cast<sockaddr *>(&saddr4);
      saddrLen = sizeof(saddr4);
    }
    if (bind(lsock, saddrP, saddrLen) < 0)
      Errorm("bind");
    if (listen(lsock, 128) < 0) Errorm("listen");

    // Set for non-blocking IO
    setSockNonblocking(lsock);

    // Tell updateSocksize that we have adjusted the size
    thLock(&globalMutex);
    currsize = socksize;
    thBcast(&globalCond);
    thUnlock(&globalMutex);

    // Loop, accepting connections
    while (true)
    {
      // If socket buffer size changes, close the listen socket and break
      // out of accept loop to go back and re-open socket with correct size
      // (since size must be set before listen call).
      if (socksize != currsize)
      {
        if (close(lsock) < 0) Errorm("close lsock");
        break;
      }

      // Wait for a connection.  Use a timeout on poll, so that if we miss
      // the signal for a socksize change, we'll wake up soon and notice
      // it.  Using ppoll would be better, but not all systems support that.
      ps.fd = lsock;
      ps.events = POLLIN;
      ps.revents = 0;
      rc = poll(&ps, 1, 1000);
      if (rc < 0)
      {
        if (errno == EINTR)
          continue;
        Errorm("poll for incoming connection");
      }
      if (rc == 0)
        continue;

      // Accept anybody who connects
      alen = saddrLen;
      sock = accept(lsock, saddrP, &alen);
      if (sock == INVALID_SOCK)
      {
        if (errno == EAGAIN || errno == EWOULDBLOCK ||
            errno == ECONNABORTED || errno == EINTR)
          continue;
        Errorm("accept");
      }

      // Get peer address
      alen = saddrLen;
      if (getpeername(sock, saddrP, &alen) < 0)
        Errorm("getpeername");
      iaddr.loadSockaddr(saddrP);
      Log("Connection from " << iaddr.tostring());

      // Don't delay sending messages
      if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on)) < 0)
        Errorm("setsockopt TCP_NODELAY");

      // Set for non-blocking IO
      setSockNonblocking(sock);

      // Hand off the socket to a receiver thread
      pickReceiver()->addConn(sock, new TcpConn(sock, iaddr));
    }
  }
  return 0;
}


// Tell the listen/accept thread that the socket send/receive buffer size
// has changed, so the listen socket must be closed and re-opened to set
// the new size.
void ListenAccept::updateSocksize()
{
  if (socksize == currsize)
    return;

  // Wake up the thread, if it is in poll.  If it is just about to start
  // the poll, it could miss this signal, but it will time out soon after
  // that and notice the change.
  thKill(getThread(), SIGUSR1);

  // Wait until the thread has picked up the new size.
  thLock(&globalMutex);
  while (socksize != currsize)
    thWait(&globalCond, &globalMutex);
  thUnlock(&globalMutex);
}


#ifdef RDMA
// Handle a work completion event
static void handleCQEvent(struct ibv_wc *wcP)
{
  PollWait *pwaitP = reinterpret_cast<PollWait *>(wcP->wr_id);

  if (wcP->status != IBV_WC_SUCCESS)
  {
    if (debugLevel > 1)
      Logt("RDMA event, qp " << wcP->qp_num << " pwait " << pwaitP
           << " status " << wcP->status);
    pwaitP->wakeup(wcP->status);
    return;
  }

  if (debugLevel > 1)
    Logt("RDMA event, qp " << wcP->qp_num << " pwait " << pwaitP
         << " op " << wcP->opcode);

  switch(wcP->opcode)
  {
    case IBV_WC_SEND:
    case IBV_WC_RDMA_WRITE:
    case IBV_WC_RDMA_READ:
      pwaitP->wakeup(wcP->status);
      break;

    case IBV_WC_RECV:
      if (pwaitP->connP == NULL)
        Error("RDMA receive complete with no connection");
      pwaitP->connP->rdmaRecv(pwaitP, wcP->byte_len);
      break;

    default:
      Error("Invalid opcode " << wcP->opcode << " in WC");
  }
}


// RDMA receiver thread body
int RdmaReceiver::threadBody()
{
  int j, rc, nEvents = 0, nEntries = cqSize;
  struct ibv_wc *wcP;
  struct ibv_cq *ecq;
  void *ecqContext;

  if (nEntries < 1)
    Error("bad CQ size");
  wcP = new ibv_wc[nEntries];

  // Wait for events on RDMA completion channel and dispatch them.
  // Exit if rdmaRcvRun is turned off.
  while (rdmaRcvRun)
  {
    // Request completion notifications
    if (ibv_req_notify_cq(ibCQ, 0) != 0)
      Errorm("ibv_req_notify_cq failed");

    // Grab all entries out of the completion queue.  We do this after
    // requesting notifications but before calling ibv_get_cq_event because
    // otherwise we might miss an entry that was added before notifications
    // were enabled.
    while (true)
    {
      rc = ibv_poll_cq(ibCQ, nEntries, wcP);
      if (rc == 0)
        break;
      if (rc < 0)
        Error("ibv_poll_cq failed");
      if (rc > nEntries)
        Error("ibv_poll_cq returned too many entries");
      for (j = 0; j < rc; j++)
        handleCQEvent(&wcP[j]);
    }
    
    // Wait for an event on completion channel.  If someone wants us to go
    // away, we could be interrupted here by a SIGUSR1, so don't die if
    // that happens.
    rc = ibv_get_cq_event(ibCC, &ecq, &ecqContext);
    if (rc == 0)
    {
      nEvents++;
      if (ecq != ibCQ)
        Error("ibv_get_cq_event returned unknown CQ");
    }
    else if (rdmaRcvRun)
      Error("ibv_get_cq_event failed");
  }

  // Acknowledge all the events we received so that CQ can be deleted
  if (nEvents > 0)
    ibv_ack_cq_events(ibCQ, nEvents);

  // Notify waitForDeath callers that we are done
  thLock(&globalMutex);
  running = false;
  thBcast(&globalCond);
  thUnlock(&globalMutex);

  delete [] wcP;
  if (debugLevel > 0)
    Logt("RdmaReceiver thread terminated");
  return 0;
}


// Convert RDMA async event type to string
string rdmaEventToStr(enum ibv_event_type ev)
{
  switch (ev)
  {
    case IBV_EVENT_QP_FATAL:            return "IBV_EVENT_QP_FATAL";
    case IBV_EVENT_QP_REQ_ERR:          return "IBV_EVENT_QP_REQ_ERR";
    case IBV_EVENT_QP_ACCESS_ERR:       return "IBV_EVENT_QP_ACCESS_ERR";
    case IBV_EVENT_COMM_EST:            return "IBV_EVENT_COMM_EST";
    case IBV_EVENT_SQ_DRAINED:          return "IBV_EVENT_SQ_DRAINED";
    case IBV_EVENT_PATH_MIG:            return "IBV_EVENT_PATH_MIG";
    case IBV_EVENT_PATH_MIG_ERR:        return "IBV_EVENT_PATH_MIG_ERR";
    case IBV_EVENT_QP_LAST_WQE_REACHED: return "IBV_EVENT_QP_LAST_WQE_REACHED";
    case IBV_EVENT_CQ_ERR:              return "IBV_EVENT_CQ_ERR";
    case IBV_EVENT_SRQ_ERR:             return "IBV_EVENT_SRQ_ERR";
    case IBV_EVENT_SRQ_LIMIT_REACHED:   return "IBV_EVENT_SRQ_LIMIT_REACHED";
    case IBV_EVENT_PORT_ACTIVE:         return "IBV_EVENT_PORT_ACTIVE";
    case IBV_EVENT_PORT_ERR:            return "IBV_EVENT_PORT_ERR";
    case IBV_EVENT_LID_CHANGE:          return "IBV_EVENT_LID_CHANGE";
    case IBV_EVENT_PKEY_CHANGE:         return "IBV_EVENT_PKEY_CHANGE";
    case IBV_EVENT_SM_CHANGE:           return "IBV_EVENT_SM_CHANGE";
    case IBV_EVENT_CLIENT_REREGISTER:   return "IBV_EVENT_CLIENT_REREGISTER";
    case IBV_EVENT_DEVICE_FATAL:        return "IBV_EVENT_DEVICE_FATAL";
    default:                            return "Unknown";
  }
  return "?";
}

   
// RDMA async event handler thread body
int RdmaAsync::threadBody()
{
  struct ibv_async_event event;
  ostringstream os;

  while (rdmaAsyncRun)
  {
    if (ibv_get_async_event(ibContext, &event) != 0)
    {
      if (!rdmaAsyncRun)
        break;
      Error("ibv_get_async_event failed");
    }

    os.str("RDMA async event ");
    os << rdmaEventToStr(event.event_type);

    switch (event.event_type)
    {
      // QP events
      case IBV_EVENT_QP_FATAL:
      case IBV_EVENT_QP_REQ_ERR:
      case IBV_EVENT_QP_ACCESS_ERR:
      case IBV_EVENT_COMM_EST:
      case IBV_EVENT_SQ_DRAINED:
      case IBV_EVENT_PATH_MIG:
      case IBV_EVENT_PATH_MIG_ERR:
      case IBV_EVENT_QP_LAST_WQE_REACHED:
        os << " on QP " << event.element.qp;
        break;

      // CQ events
      case IBV_EVENT_CQ_ERR:
        os << " on CQ " << event.element.cq;
        break;

      // SRQ events
      case IBV_EVENT_SRQ_ERR:
      case IBV_EVENT_SRQ_LIMIT_REACHED:
        os << " on SRQ " << event.element.srq;
        break;

      // Port events
      case IBV_EVENT_PORT_ACTIVE:
      case IBV_EVENT_PORT_ERR:
      case IBV_EVENT_LID_CHANGE:
      case IBV_EVENT_PKEY_CHANGE:
      case IBV_EVENT_SM_CHANGE:
      case IBV_EVENT_CLIENT_REREGISTER:
        os << " on port " << event.element.port_num;
        break;

      // CA events
      case IBV_EVENT_DEVICE_FATAL:
        break;

      default:
        break;
    }
    Warn(os.str());
    ibv_ack_async_event(&event);
  }

  // Notify waitForDeath callers that we are done
  thLock(&globalMutex);
  running = false;
  thBcast(&globalCond);
  thUnlock(&globalMutex);

  if (debugLevel > 0)
    Logt("RdmaAsync thread terminated");
  return 0;
}
#endif // RDMA


// Print help message
static void helpCmd(vector<string> *argsP)
{
  cout << "Commands:"
       << endl
       << "  help                  Print help message" << endl
       << "  quit                  Exit from program" << endl
       << "  version               Show program version" << endl
       << "  debug [LEVEL]         Set debugging output level" << endl
       << "  source FILENAME...    Read commands from file" << endl
       << "  check HOSTNAME...     Quick performance test to single nodes from local" << endl
       << "  server HOSTNAME...    Designate nodes as servers" << endl
       << "  client HOSTNAME...    Designate nodes as clients" << endl
       << "  delete HOSTNAME...    Remove server or client nodes from test" << endl
       << "  reset                 Close connections and clear node tables" << endl
       << "  status                Show status of all nodes" << endl
       << "  ttime NSEC            Set run time (in seconds) for tests" << endl
       << "  buffsize NBYTES       Set I/O data buffer size for tests" << endl
       << "  socksize NBYTES       Set size of TCP send/receive buffer space" << endl
       << "  threads NTHREADS      Set number of tester threads to use on clients" << endl
       << "                          (default " << TESTER_THREADS << ")" << endl
       << "  parallel N            Set number of parallel socket connections (default 1)" << endl
       << "  rdma [on|off|all|inline] Enable or disable RDMA for sending data blocks" << endl
       << "  hist [on|off]         Set printing of response time histograms on or off" << endl
       << "                          (if no option given, then toggle)" << endl
       << "  verify [on|off]       Verify that contents of data messages are correct" << endl
       << "  plot [FILENAME]       Write test results in gnuplot format to specified file" << endl
       << "                          (if no filename given, then turn off plotting)" << endl
       << "  test TTYPE...         Run performance tests from clients to servers" << endl
       << "  kill HOSTNAME...      Shut down " << progname << " server on specified nodes" << endl
       << "  killall               Shut down " << progname << " on all client and server nodes" << endl
       << endl
       << "Test types:" << endl
       << "  write         Clients write round-robin to all servers" << endl
       << "  read          Clients read round-robin from all servers" << endl
       << "  nwrite        Same as write test, but use NSD-style writing" << endl
       << "  swrite        Each tester thread writes to only one server" << endl
       << "  sread         Each tester thread reads from only one server" << endl
       << "  rw            Half of the tester threads read and half write" << endl;
}


// Exit from program
static void quitCmd(vector<string> *argsP)
{
  quitflag = true;
}


// Read commands from file
static void sourceCmd(vector<string> *argsP)
{
  vector<string>::iterator arg;
  ifstream infile;
  string fname, line;

  if (argsP->empty())
  {
    Log("File name missing");
    return;
  }
  nestingLevel++;
  if (nestingLevel > 10)
  {
    Log("Nesting level too deep");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    fname = *arg;
    infile.open(fname.c_str());
    if (!infile)
    {
      Logm("open of " << fname << " failed");
      break;
    }
    while (getline(infile, line))
      splitCmd(line, &commands);
    infile.close();
  }
}


// Show program version
static void versionCmd(vector<string> *argsP)
{
  Log(version);
}


// Set debugging output level
static void debugCmd(vector<string> *argsP)
{
  if (argsP->empty())
    debugLevel++;
  else
    debugLevel = atoi((*argsP)[0].c_str());
  Log("Debugging output level set to " << debugLevel);
}


// Send current test parameters to target.  Return true if error occurred.
static bool sendParms(Target *targP)
{
  DataBuff db(8 * sizeof(UInt32));
  db.putUInt32(testTime);
  db.putUInt32(buffsize);
  db.putUInt32(socksize);
  db.putUInt32(nTesterThreads);
  db.putUInt32(nParallel);
  db.putUInt32(useRdma);
  db.putUInt32(verify);
  db.putUInt32(clientNodes.size());

  RcvMsg *rmsgP = targP->sendm(mtParms, &db);
  bool gotErr = rmsgP->showError();
  delete rmsgP;
  return gotErr;
}


// Quick single-threaded performance test to individual nodes from local
static void checkCmd(vector<string> *argsP)
{
  DataBuff wbuff, rbuff;
  IpAddr iaddr;
  Target *targP;
  string errmsg;
  RcvMsg *rmsgP;
  HTime startTime, endTime, lastTime;
  UInt64 totBytes;
  vector<string>::iterator arg;
  Histogram hist;
  bool gotError;
  list<TType> tests;
  list<TType>::iterator tt;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }
  if (useRdma != rOff)
  {
    Log("Check command isn't supported with RDMA enabled.");
    return;
  }

  wbuff.newBuff(buffsize);
  wbuff.fillBuff(randSeed());

  rbuff.newBuff(sizeof(UInt64) + sizeof(RdmaAddr) + sizeof(UInt32));
  rbuff.putUInt64(randSeed());
  rbuff.putRdmaAddr(RdmaAddr());
  rbuff.putUInt32(0);

  tests.push_back(ttWrite);
  tests.push_back(ttRead);

  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    if (iaddr.parse(*arg) != E_OK)
      continue;
    targP = new Target(*arg, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
    {
      Log(errmsg);
      delete targP;
      continue;
    }

    // Tell remote node the test parameters
    if (sendParms(targP))
      return;

    for (tt = tests.begin(); tt != tests.end(); ++tt)
    {
      totBytes = 0;
      startTime = endTime = getTime();
      gotError = false;
      while (endTime - startTime < sectoht(testTime))
      {
        lastTime = endTime;

        if (*tt == ttWrite)
          rmsgP = targP->sendm(mtWrite, &wbuff);
        else
          rmsgP = targP->sendm(mtRead, &rbuff);
        if (rmsgP->showError())
          gotError = true;
        delete rmsgP;
        if (gotError)
          break;

        endTime = getTime();
        hist.addEntry(endTime - lastTime);
        totBytes += buffsize + MSG_HDRSIZE;
      }
      if (gotError)
      {
        delete targP;
        continue;
      }

      double elapsedT = httosec(endTime - startTime);
      double rate = siground(totBytes / elapsedT / 1000000.0, 3);
      if (*tt == ttWrite)
        Log("write rate " << rate << " MB/sec");
      else
        Log("read rate " << rate << " MB/sec");
      if (showHist)
        cout << "block transmit times (average "
             << siground(hist.average() * 1000.0, 4) << " msec, median "
             << siground(hist.median() * 1000.0, 4) << " msec)" << endl
             << "     msec  nevents" << endl
             << hist;
    }
    delete targP;
  }
}


// Close connections and clear node tables
static void resetCmd(vector<string> *argsP)
{
  map<IpAddr, Target *>::iterator node;

  // Tell every node that we know about to close any connections that
  // it has, and then remove the node from local tables.
  for (node = allNodes.begin(); node != allNodes.end(); )
  {
    Target *targP = node->second;
    RcvMsg *rmsgP = targP->sendm(mtReset);
    rmsgP->showError();
    delete rmsgP;
    delete targP;
    allNodes.erase(node++);
  }
  serverNodes.clear();
  clientNodes.clear();
}


// Shut down all client and server nodes
static void killallCmd(vector<string> *argsP)
{
  map<IpAddr, Target *>::iterator node;
  for (node = allNodes.begin(); node != allNodes.end(); )
  {
    Target *targP = node->second;
    delete targP->sendm(mtKill);
    delete targP;
    allNodes.erase(node++);
  }
  serverNodes.clear();
  clientNodes.clear();
}


// Shut down server on specified nodes
static void killCmd(vector<string> *argsP)
{
  IpAddr iaddr;
  Target *targP;
  string errmsg;
  vector<string>::iterator arg;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    if (iaddr.parse(*arg) != E_OK)
      continue;
    targP = new Target(*arg, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
      Log("Cannot connect to " << *arg << ": " << errmsg);
    else
      delete targP->sendm(mtKill);
    delete targP;
  }
}


// Common routine for serverCmd and clientCmd
static void doClientServer(vector<string> *argsP, list<Target *> *targListP)
{
  string hname, errmsg;
  IpAddr iaddr;
  Target *targP;
  vector<string>::iterator arg;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    hname = *arg;
    if (iaddr.parse(hname) != E_OK)
      continue;
    if (allNodes.find(iaddr) != allNodes.end())
    {
      Log(hostString(hname, iaddr) << " is already in use");
      continue;
    }
    targP = new Target(hname, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
    {
      Log("Cannot connect to " << hname << ": " << errmsg);
      delete targP;
    }
    else
      targListP->push_back(targP);
  }
}


// Designate nodes as servers
static void serverCmd(vector<string> *argsP)
{
  list<Target *> targList;
  doClientServer(argsP, &targList);

  list<Target *>::const_iterator targ;
  for (targ = targList.begin(); targ != targList.end(); ++targ)
  {
    serverNodes.insert(pair<IpAddr, Target *>((*targ)->iaddr, *targ));
    allNodes[(*targ)->iaddr] = *targ;
  }
}


// Designate nodes as clients
static void clientCmd(vector<string> *argsP)
{
  list<Target *> targList;
  doClientServer(argsP, &targList);

  list<Target *>::const_iterator targ;
  for (targ = targList.begin(); targ != targList.end(); ++targ)
  {
    clientNodes[(*targ)->iaddr] = *targ;
    allNodes[(*targ)->iaddr] = *targ;
  }
}


// Remove server or client nodes from test
static void deleteCmd(vector<string> *argsP)
{
  string hname;
  IpAddr iaddr;
  Target *targP;
  vector<string>::iterator arg;
  map<IpAddr, Target *>::iterator node;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    hname = *arg;
    if (iaddr.parse(hname) != E_OK)
      return;
    node = allNodes.find(iaddr);
    if (node == allNodes.end())
    {
      Log(hostString(hname, iaddr) << " not found");
      continue;
    }
    targP = node->second;

    // Don't worry if send fails.  If we ever use this node again, we'll
    // send another mtReset and check results from that.
    if (targP->connP != NULL)
      delete targP->sendm(mtReset);

    delete targP;
    allNodes.erase(node);
    serverNodes.erase(iaddr);
    clientNodes.erase(iaddr);
  }
}


// Show status of a target
static void showStatus(pair<IpAddr, Target *> p)
{
  Target *targP = p.second;
  cout << "  " << targP->name();
  if (targP->connP == NULL)
  {
    cout << ", not connected" << endl;
    return;
  }
  cout << endl;

  RcvMsg *rmsgP = targP->sendm(mtStatus);
  if (!rmsgP->errText.empty())
  {
    Log("    " << rmsgP->errText);
    delete rmsgP;
    return;
  }
  cout << rmsgP->msgBuff.getString();
  delete rmsgP;
}


// If the TCP connection for this target is not connected, try to reconnect it
static void reconnect(pair<IpAddr, Target *> p)
{
  Target *targP = p.second;
  if (targP->connP != NULL)
  {
    if (!targP->connP->isBroken())
      return;
    targP->connP->connShutdown();
    targP->connP->releaseConn();
    targP->connP = NULL;
  }
  (void) targP->makeConnection();
}


// Show status of all nodes
static void statusCmd(vector<string> *argsP)
{
  string r;
  switch (useRdma)
  {
    case rOff:    r = "no";     break;
    case rOn:     r = "yes";    break;
    case rAll:    r = "all";    break;
    case rInline: r = "inline"; break;
  }
  string v = verify ? "on" : "off";
  cout << "test time: " << testTime << " sec" << endl
       << "data buffer size: " << buffsize << endl
       << "TCP socket send/receive buffer size: " << socksize << endl
       << "tester threads: " << nTesterThreads << endl
       << "parallel connections: " << nParallel << endl
       << "RDMA enabled: " << r << endl
       << "message verification: " << v << endl
       << endl;

  if (!clientNodes.empty())
  {
    Log("clients:");
    for_each(clientNodes.begin(), clientNodes.end(), reconnect);
    for_each(clientNodes.begin(), clientNodes.end(), showStatus);
  }
  if (!serverNodes.empty())
  {
    Log("servers:");
    for_each(serverNodes.begin(), serverNodes.end(), reconnect);
    for_each(serverNodes.begin(), serverNodes.end(), showStatus);
  }
}


// Set run time for tests
static void ttimeCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Number of seconds argument missing");
    return;
  }
  testTime = atoi((*argsP)[0].c_str());
  if (testTime < 1)
    testTime = 1;
  Log("Test time set to " << testTime << " seconds");
}


// Set buffer size for tests
static void buffsizeCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Buffer size argument missing");
    return;
  }
  buffsize = atoi((*argsP)[0].c_str());
  if (buffsize < 1) buffsize = 1;
  if (buffsize > MAX_BUFFSIZE) buffsize = MAX_BUFFSIZE;
  Log("Buffer size set to " << buffsize << " bytes");
}


// Set size of TCP send/receive buffer space
static void socksizeCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Size argument missing");
    return;
  }
  socksize = atoi((*argsP)[0].c_str());
  if (socksize < 0) socksize = 0;
  if (socksize > MAX_SOCKSIZE) socksize = MAX_SOCKSIZE;
  Log("TCP send/receive buffer size set to " << socksize << " bytes");
}


// Set number of tester threads to use on clients
static void threadsCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Number of threads argument missing");
    return;
  }
  nTesterThreads = atoi((*argsP)[0].c_str());
  if (nTesterThreads < 1)   nTesterThreads = 1;
  if (nTesterThreads > MAX_TESTERS) nTesterThreads = MAX_TESTERS;
  Log("Number of tester threads set to " << nTesterThreads);
}


// Tell clients to connect to servers
static Errno doConnect()
{
  Errno err = E_OK;
  map<IpAddr, Target *>::const_iterator node;
  multimap<IpAddr, Target *>::const_iterator snode;
  MsgRecord mr;
  RcvMsg *rmsgP;
  DataBuff db;
  bool gotError;

  if (serverNodes.empty())
  {
    Log("No server nodes found");
    return E_INVAL;
  }
  if (clientNodes.empty())
  {
    Log("No client nodes found");
    return E_INVAL;
  }
  
  // Tell the server nodes to close any connections that they might have
  // had from previous sessions, where they were clients.
  for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
    if (snode->second->connP->sendMessage(mtReset, NULL, &mr, NULL) != E_OK)
    {
      Log("Send to " << snode->second->hostname << " failed");
      err = E_SENDFAILED;
    }
  if (mr.checkReplies())
    return E_REPLY;

  // Make a list of all of the servers to send to the clients
  int len = sizeof(UInt32);
  for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
    len += calcLen(snode->second->hostname) + sizeof(UInt32)
      + sizeof(snode->second->iaddr.a);
  db.newBuff(len);
  db.putUInt32(serverNodes.size());
  for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
  {
    db.putString(snode->second->hostname);
    db.putIpAddr(snode->second->iaddr);
  }

  // Send the server list to all clients
  gotError = false;
  for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
    if (node->second->connP->sendMessage(mtConnect, &db, &mr, NULL) != E_OK)
    {
      Log("Send to " << node->second->hostname << " failed");
      gotError = true;
    }
  mr.waitForReplies();

  while (true)
  {
    rmsgP = mr.nextReply();
    if (rmsgP == NULL)
      break;
    if (rmsgP->showError())
      gotError = true;
    else if (debugLevel > 0)
      Log(rmsgP->connP->destName() << " connected");
    delete rmsgP;
  }
  return gotError ? E_CONNFAILED : E_OK;
}


// Run performance test from clients to servers
static void testCmd(vector<string> *argsP)
{
  vector<string>::iterator arg;
  map<IpAddr, Target *>::iterator node;
  multimap<IpAddr, Target *>::const_iterator snode;
  list<TType> tests;
  list<TType>::iterator tt;
  MsgRecord mr, mrstat;
  HTime startTime, endTime, runTime, totTime;
  RcvMsg *rmsgP = NULL;
  Target *targP;
  DataBuff db;
  bool gotIdle, gotError;
  UInt32 idlePct, totMsgs;
  UInt64 totBytes, clientIdle, serverIdle;
  Histogram *hP;
  ofstream plfile;
  bool plotOpened = false;

  if (argsP->empty())
  {
    tests.push_back(ttWrite);
    tests.push_back(ttRead);
  }
  else
  {
    for (arg = argsP->begin(); arg != argsP->end(); ++arg)
    {
      if (match("write", *arg))
        tests.push_back(ttWrite);
      else if (match("read", *arg))
        tests.push_back(ttRead);
      else if (match("nwrite", *arg))
        tests.push_back(ttNwrite);
      else if (match("rw", *arg))
      {
        if (nTesterThreads < 2)
        {
          Log("Need at least two tester threads for rw test");
          return;
        }
        tests.push_back(ttRW);
      }
      else if (match("swrite", *arg))
        tests.push_back(ttSwrite);
      else if (match("sread", *arg))
        tests.push_back(ttSread);
      else
      {
        Log("Invalid test type: " << *arg);
        Log("Valid types are: write nwrite read rw swrite sread");
        return;
      }
    }
  }

  if (clientNodes.empty())
  {
    Log("No client nodes found");
    return;
  }
  for_each(allNodes.begin(), allNodes.end(), reconnect);
  for (node = allNodes.begin(); node != allNodes.end(); ++node)
    if (node->second->connP == NULL)
    {
      Log(node->second->name() << " is not connected");
      return;
    }

  // Tell everyone the current test parameters
  gotError = false;
  for (node = allNodes.begin(); node != allNodes.end(); ++node)
    if (sendParms(node->second))
      gotError = true;
  if (gotError)
    return;

  // Pass list of servers to all clients and have them connect
  if (doConnect() != E_OK)
    return;

  if (!plotFname.empty())
  {
    plfile.open(plotFname.c_str());
    if (!plfile)
    {
      Logm("open of " << plotFname << " failed");
      goto exit;
    }
    plotOpened = true;
  }

  db.newBuff(sizeof(UInt32));
  for (tt = tests.begin(); tt != tests.end(); ++tt)
  {
    startTime = getTime();
    runTime = sectoht(testTime);

    // Tell all clients to start the test, but don't start collecting
    // statistics yet.
    db.resetBuff();
    db.putUInt32(*tt);

    for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      if (node->second->connP->sendMessage(mtTest, &db, &mr, NULL) != E_OK)
      {
        Log("Send to " << node->second->name() << " failed");
        goto exit;
      }

    // Delay for 10% of test time and then tell all nodes to start
    // gathering statistics.
    totTime = sleepUntil(startTime + runTime/10);
    for (node = allNodes.begin(); node != allNodes.end(); ++node)
    {
      targP = node->second;
      if (targP->connP->sendMessage(mtStatOn, NULL, &mrstat, NULL) != E_OK)
      {
        Log("Send to " << targP->name() << " failed");
        goto exit;
      }
    }
    if (mrstat.checkReplies())
      goto exit;

    // Stop gathering statistics after 90% of test time.  Verify that we
    // didn't use all of the test time getting started.  We need at least
    // 100 msec left over to run the test
    endTime = startTime + runTime - runTime/10;
    if (httomsec(endTime - getTime()) < 100)
    {
      Log("Test startup took too long");
      goto exit;
    }
    sleepUntil(endTime);
    for (node = allNodes.begin(); node != allNodes.end(); ++node)
    {
      targP = node->second;
      if (targP->connP->sendMessage(mtStatOff, NULL, &mrstat, NULL) != E_OK)
      {
        Log("Send to " << targP->name() << " failed");
        goto exit;
      }
    }
    if (mrstat.checkReplies())
      goto exit;
    totTime = getTime() - totTime;

    // Wait for all clients to finish testing
    mr.waitForReplies();

    // Add up test results
    totBytes = totMsgs = clientIdle = 0;
    gotIdle = true;
    while (true)
    {
      rmsgP = mr.nextReply();
      if (rmsgP == NULL)
        break;
      if (rmsgP->showError())
        goto exit;

      totBytes += rmsgP->msgBuff.getUInt64();
      idlePct = rmsgP->msgBuff.getUInt32();
      if (idlePct <= 100)
        clientIdle += idlePct;
      else
        gotIdle = false;
      rmsgP->connP->getHistP()->getBuff(&rmsgP->msgBuff);
      totMsgs += rmsgP->connP->getHistP()->getNevents();
      delete rmsgP;
    }

    // Gather server CPU idle percentage
    multimap<IpAddr, Target *>::const_iterator snode;
    for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
      if (snode->second->connP->sendMessage(mtIdlePct, NULL, &mr, NULL) != E_OK)
      {
        Log("Send to " << snode->second->name() << " failed");
        goto exit;
      }
    mr.waitForReplies();

    serverIdle = 0;
    while (true)
    {
      rmsgP = mr.nextReply();
      if (rmsgP == NULL)
        break;
      if (rmsgP->showError())
        goto exit;

      idlePct = rmsgP->msgBuff.getUInt32();
      if (idlePct <= 100)
        serverIdle += idlePct;
      else
        gotIdle = false;
      delete rmsgP;
    }

    // Print results
    cout << clientNodes.size() << "-" << serverNodes.size() << " ";
    switch (*tt)
    {
      case ttWrite:  cout << "write";  break;
      case ttNwrite: cout << "nwrite"; break;
      case ttRead:   cout << "read";   break;
      case ttRW:     cout << "rw";     break;
      case ttSwrite: cout << "swrite"; break;
      case ttSread:  cout << "sread";  break;
      default:       Error("invalid test type"); break;
    }
    double elapsedT = httosec(totTime);
    double rate = siground(totBytes / elapsedT / 1000000.0, 3);
    double mrate = siground(totMsgs / elapsedT, 3);
    cout << " " << rate << " MB/sec";
    cout << " (" << mrate << " msg/sec)";
    if (gotIdle)
    {
      cout << ", cli " << 100 - clientIdle / clientNodes.size() << "%";
      cout << " srv " << 100 - serverIdle / serverNodes.size() << "%";
    }
    cout << ", time " << testTime << ", buff " << buffsize;
    if (socksize != 0)
      cout << ", sock " << socksize;
    if (nTesterThreads != TESTER_THREADS)
      cout << ", th " << nTesterThreads;
    if (nParallel > 1)
      cout << ", parallel " << nParallel;
    if (useRdma != rOff)
      cout << ", RDMA";
    if (useRdma == rAll)
      cout << " all";
    else if (useRdma == rInline)
      cout << " inline";
    if (verify)
      cout << ", verify";
    cout << endl;

    if (showHist)
    {
      cout << endl;
      for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      {
        Histogram *hP = node->second->connP->getHistP();
        cout << node->second->hostname << " block transmit times (average "
             << siground(hP->average() * 1000.0, 4) << " msec, median "
             << siground(hP->median() * 1000.0, 4) << " msec)" << endl
             << "     msec  nevents" << endl
             << *hP << endl;
      }
    }

    if (!plotFname.empty())
    {
      unsigned int n = clientNodes.size();
      if (n > 1)
      {
        int rows, cols;
        rows = (n + 3) / 4;
        cols = n > 4 ? 4 : n;
        plfile << "set multiplot layout " << rows << "," << cols
               << " title \"Block transmit times (msec)\"" << endl
               << "set xlabel" << endl
               << "set ylabel" << endl;
      }
      else
        plfile << "set xlabel \"Time (msec)\"" << endl
               << "set ylabel" << endl;

      plfile << "set format y \"\"" << endl
             << "set style data boxes" << endl
             << "set style fill solid" << endl
             << "set boxwidth 0.9" << endl
             << "unset key" << endl
             << "set border 1" << endl
             << "set xtics out nomirror offset 0,0.5" << endl
             << "unset ytics" << endl;

      // Compute bounds for all nodes and use that to set xrange and yrange
      // so that all plots will have the same scale.
      UInt32 v, maxval = 0;
      double t, minT = 9999999999999999.0, maxT = 0.0;
      for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      {
        hP = node->second->connP->getHistP();
        v = hP->maxBucket(); if (v > maxval) maxval = v;
        t = hP->minTime();   if (t < minT)   minT = t;
        t = hP->maxTime();   if (t > maxT)   maxT = t;
      }

      // Convert time bounds to next msec
      minT *= 1000.0;
      maxT *= 1000.0;
      minT = (minT < 1) ? 0 : floor(minT - 1);
      maxT = floor(maxT + 1);
      maxval = (maxval + 9) / 10 * 10;

      plfile << "set xrange [" << minT << ":" << maxT << "]" << endl
             << "set yrange [0:" << maxval << "]" << endl;

      for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      {
        hP = node->second->connP->getHistP();
        plfile << "set label \"" << node->second->hostname << "\""
               << " at graph 0.5,0.8 center" << endl
               << "plot '-' using 1:2" << endl
               << *hP << "e" << endl
               << "unset label" << endl;
      }
      if (clientNodes.size() > 1)
        plfile << "unset multiplot" << endl;

      plfile << "pause mouse any keypress" << endl;
    }
  }

  // Tear down RDMA connections
  if (useRdma != rOff)
  {
    for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
    {
      targP = node->second;
      if (targP->connP->sendMessage(mtRdmaDisconn, NULL, &mr, NULL) != E_OK)
        Log("Send to " << targP->name() << " failed");
    }
    mr.checkReplies();
  }

exit:
  // If we broke out early due to an error, clean out any outstanding
  // replies.
  delete rmsgP;
  mrstat.checkReplies();
  mr.checkReplies();
  if (plotOpened)
  {
    plfile.close();
    if (!plfile)
      Logm("close of " << plotFname << " failed");
  }
}


// Toggle printing of client resonse time histograms
static void histCmd(vector<string> *argsP)
{
  if (argsP->empty())
    showHist = !showHist;
  else if ((*argsP)[0] == "on")
    showHist = true;
  else if ((*argsP)[0] == "off")
    showHist = false;
  else
  {
    Log("Invalid option");
    return;
  }
  cout << "Histogram printing is now ";
  if (showHist) cout << "on"; else cout << "off";
  cout << endl;
}


// Set number of parallel socket connections
static void parallelCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Number of connections argument is missing");
    return;
  }
  nParallel = atoi((*argsP)[0].c_str());
  if (nParallel < 1) nParallel = 1;
  if (nParallel > MAX_PARALLEL) nParallel = MAX_PARALLEL;
  Log("Number of parallel socket connections set to " << nParallel);
}


// Use RDMA for sending data blocks
static void rdmaCmd(vector<string> *argsP)
{
#ifdef RDMA
  if (argsP->empty())
    useRdma = (useRdma != rOff) ? rOff : rOn;
  else if ((*argsP)[0] == "on")
    useRdma = rOn;
  else if ((*argsP)[0] == "off")
    useRdma = rOff;
  else if (match("all",(*argsP)[0]))
    useRdma = rAll;
  else if (match("inline", (*argsP)[0]))
    useRdma = rInline;
  else
  {
    Log("Invalid option");
    return;
  }
  string r;
  switch (useRdma)
  {
    case rOff:    r = "off";       break;
    case rOn:     r = "on";        break;
    case rAll:    r = "on all";    break;
    case rInline: r = "on inline"; break;
  }
  cout << "RDMA is now " << r << endl;
#else
  Log("RDMA is not supported");
#endif
}


// Toggle verification of data message contents
static void verifyCmd(vector<string> *argsP)
{
  if (argsP->empty())
    verify = !verify;
  else if ((*argsP)[0] == "on")
    verify = true;
  else if ((*argsP)[0] == "off")
    verify = false;
  else
  {
    Log("Invalid option");
    return;
  }
  cout << "Message verification is now ";
  if (verify) cout << "on"; else cout << "off";
  cout << endl;
}


// Write test results in gnuplot format to specified file
static void plotCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    plotFname.clear();
    Log("Plotting disabled");
    return;
  }
  plotFname = (*argsP)[0];
  Log("Plot commands will be written to " << plotFname);
}


// Entry in command table
struct Command
{
  string cmd;                                   // Command keyword
  string::size_type minlen;                     // Minium abbreviation
  void (*cmdRtnP)(vector<string> *argsP);       // Routine to call
};

  
// Valid commands
static struct Command cmdTable[] = 
{
  { "hist",     2, histCmd },
  { "help",     1, helpCmd },
  { "?",        1, helpCmd },
  { "quit",     1, quitCmd },
  { "source",   2, sourceCmd },
  { ".",        1, sourceCmd },
  { "version",  1, versionCmd },
  { "debug",    3, debugCmd },
  { "check",    2, checkCmd },
  { "reset",    1, resetCmd },
  { "killall",  5, killallCmd },
  { "ka",       2, killallCmd },
  { "kill",     1, killCmd },
  { "status",   2, statusCmd },
  { "socksize", 2, socksizeCmd },
  { "server",   1, serverCmd },
  { "client",   1, clientCmd },
  { "delete",   1, deleteCmd },
  { "threads",  2, threadsCmd },
  { "ttime",    2, ttimeCmd },
  { "plot",     1, plotCmd },
  { "parallel", 2, parallelCmd },
  { "rdma",     2, rdmaCmd },
  { "verify",   4, verifyCmd },
  { "test",     1, testCmd },
  { "buffsize", 1, buffsizeCmd },
};
static const int nCommands = sizeof(cmdTable) / sizeof(struct Command);  


// Handle SIGUSR1
static void sigusr1Handler(int signum)
{ }


// Handle out of memory error
static void exhausted()
{
  cerr << "Out of storage!" << endl;
  exit(EXIT_FAILURE);
}


// Print usage message and exit
static void usage()
{
  cerr << "usage:  " << progname
       << " [-d] [-h] [-i FNAME] [-p PORT] [-r RDMAPORT] [-t NRCV] [-s]" << endl
       << "                [-w NWORKERS] [-6] [CMD...]" << endl
       << endl
       << "Options:" << endl
       << "  -d            Include debug output" << endl
       << "  -h            Print help message" << endl
       << "  -i FNAME      Read commands from file FNAME" << endl
       << "  -p PORT       TCP port to use (default " << NSDPERF_PORT << ")" << endl
       << "  -r RDMAPORT   RDMA device and port to use (default first device, port 1)" << endl
       << "  -t NRCV       Number of receiver threads (default nCPUs, min 2)" << endl
       << "  -s            Act as a server" << endl
       << "  -w NWORKERS   Number of message worker threads (default " << MSG_WORKERS << ")" << endl
       << "  -6            Use IPv6 rather than IPv4" << endl;

  exit(EXIT_FAILURE);
}


// Start of main program
int main(int argc, char *argvP[])
{
  int j, c;
  string::size_type spos;
  string line;

  // Trap out-of-memory errors
  set_new_handler(exhausted);

  // Get the name of this program for error messages
  progname = argvP[0];
  spos = progname.find_last_of("/\\");
  if (spos != string::npos) progname = progname.substr(spos+1);

  // Figure out the endian-ness of this machine
  union { UInt16 n; UChar c[2]; } endTest;
  endTest.n = 1;
  littleEndian = (endTest.c[0] == 1);

  // Get page size of this machine
  pagesize = sysconf(_SC_PAGE_SIZE);

  // Parse options
  string fname, rdport;
  int nReceivers = numProcessors();
  while ((c = getopt(argc, argvP, "dhi:p:r:t:sw:6")) != EOF)
    switch (c)
    {
      case 'd':                 // Include debug output
        debugLevel++;
        break;

      case 'h':                 // Ask for help
        usage();
        break;

      case 'i':                 // Input file name
        fname = optarg;
        break;

      case 'p':                 // Port to use
        port = atoi(optarg);
        break;

      case 'r':                 // RDMA device and port
        rdport = optarg;
        spos = 0;
        rdmaDev = getword(rdport, spos, ":/");
        if (spos != string::npos)
          rdmaPort = atoi(rdport.substr(spos+1).c_str());
        break;

      case 't':                 // Number of receiver threads
        nReceivers = atoi(optarg);
        break;

      case 's':                 // Run in server mode
        server = true;
        break;

      case 'w':                 // Number of message worker threads
        nWorkers = atoi(optarg);
        break;

      case '6':                 // Use IPv6 rather than IPv4
        useipv6 = true;
        break;

      default:  
        exit(EXIT_FAILURE);
        break;
    }

#ifdef IPV6_SUPPORT
  addrFamily = useipv6 ? AF_INET6 : AF_INET;
#else
  if (useipv6) Error("IPv6 not supported on this system");
  addrFamily = AF_INET;
#endif
  if (nReceivers < 2) nReceivers = 2;
  if (nReceivers > 32) nReceivers = 32;
  srandom(12345);

  // If input file name was given, read file contents into line buffer
  if (!fname.empty())
  {
    ifstream infile;
    infile.open(fname.c_str());
    if (!infile)
      Errorm("open of " << fname << " failed");
    while (getline(infile, line))
      splitCmd(line, &commands);
    infile.close();
  }

  // Remaining arguments are treated as a command line
  for (j = optind; j < argc; j++)
  {
    if (!line.empty())
      line += " ";
    line += argvP[j];
  }
  splitCmd(line, &commands);
  
  // Set up a signal handler for SIGUSR1
  struct sigaction sa;
  sa.sa_handler = sigusr1Handler;
  sa.sa_flags = 0;
  if (sigaction(SIGUSR1, &sa, NULL) < 0) Errorm("sigaction");

  // Start up threads
  thInitMutex(&logMutex);
  thInitMutex(&globalMutex);
  thInitCond(&globalCond);
  for (j = 0; j < nReceivers; j++)
  {
    Receiver *rcvP = new Receiver;
    rcvP->init();
    receiverTab.push_back(rcvP);
  }
  nextReceiver = receiverTab.begin(); 
  thInitMutex(&workerMutex);
  thInitCond(&workerCond);
  for (j = 0; j < nWorkers; j++)
  {
    MsgWorker *mwP = new MsgWorker();
    mwP->init();
    workerTab.push_back(mwP);
  }

  thInitMutex(&testerMutex);
  thInitCond(&testerCond);

  if (server)
  {
    laThreadP = new ListenAccept;
    laThreadP->init();
    if (debugLevel > 0)
      Log(progname << " server started");

    // Clean up threads if they exit, and quit when all threads are gone
    thLock(&globalMutex);
    while (nThreadsStarted > 0)
    {
      while (deadThreads.empty())
        thWait(&globalCond, &globalMutex);

      Thread *thP = deadThreads.front();
      deadThreads.pop_front();
      thUnlock(&globalMutex);

      thJoin(thP->getThread(), NULL);
      delete thP;

      thLock(&globalMutex);
      nThreadsStarted--;
    }
    thUnlock(&globalMutex);
    exit(EXIT_SUCCESS);
  }

  // Read commands
  bool interactive = isatty(fileno(stdin));
  while (true)
  {
    // If no more pre-read lines, prompt for a new one
    while (commands.empty())
    {
      nestingLevel = 0;
      if (interactive)
        cout << progname << "> ";
      if (!getline(cin, line))
      {
        if (interactive)
          cout << endl;
        quitflag = true;
        break;
      }
      splitCmd(line, &commands);
    }
    if (quitflag)
      break;

    line = commands.front();
    commands.pop_front();

    // First word in line is command name
    spos = 0;
    string cmd = getword(line, spos);
    if (cmd.empty() || match(cmd, "#"))
      continue;

    // Split remainder of line into words, which are the command arguments
    vector<string> args;
    while (spos != string::npos)
    {
      string w = getword(line, spos);
      if (!w.empty())
        args.push_back(w);
    }

    // Look up command in table and run the command routine
    for (j = 0; j < nCommands; j++)
      if (cmd.length() >= cmdTable[j].minlen && match(cmdTable[j].cmd, cmd))
      {
        cmdTable[j].cmdRtnP(&args);
        break;
      }
    if (j == nCommands)
    {
      Log("Invalid command.");
      helpCmd(NULL);
    }
    if (quitflag)
      break;
  }
  exit(EXIT_SUCCESS);
}
